import json
import math
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, List, Optional

import jax.numpy as jnp
import numpy as np
from jax.sharding import Mesh

from tpu_inference import envs, utils

if TYPE_CHECKING:
    from vllm.v1.configs.vllm_config import VllmConfig

MESH_AXIS_NAMES = ("data", "attn_dp", "expert", "model")
MESH_AXIS_NAMES_2D = ('data', 'model')


class ShardingAxisNameBase:
    """Base class for sharding axis names."""
    SEQUENCE = ('data', 'attn_dp')
    ATTN_DATA = ('data', 'attn_dp')
    MLP_DATA = 'data'
    ATTN_HEAD = 'model'
    ATTN_TENSOR = None
    MLP_TENSOR = ('attn_dp', 'model', 'expert')
    MOE_TENSOR = ('attn_dp', 'model')
    EXPERT = ('attn_dp', 'expert', 'model')
    VOCAB = ('expert', 'model')


class ShardingAxisName2D:
    """Sharding axis names for 2D data parallelism scenarios.
    NOTE(wenxindongwork): The new MoE kernel expects a 2D mesh for now.
    We should use ShardingAxisNameBase once the new MoE kernel supports
    more general mesh shapes. For now, this is the default sharding axes.
    """
    SEQUENCE = 'data'
    ATTN_DATA = 'data'
    MLP_DATA = 'data'
    ATTN_HEAD = 'model'
    ATTN_TENSOR = None
    MLP_TENSOR = 'model'
    MOE_TENSOR = 'model'
    EXPERT = 'model'
    VOCAB = ('data', 'model')


try:
    _use_base_sharding = envs.NEW_MODEL_DESIGN
    if _use_base_sharding:
        ShardingAxisName = ShardingAxisNameBase
    else:
        ShardingAxisName = ShardingAxisName2D
except Exception:
    ShardingAxisName = ShardingAxisName2D


@dataclass
class ShardingStrategy:
    """Defines the high-level parallelism strategy.

    This class specifies how many ways each type of parallelism (tensor, expert,
    sequence, data) should be distributed across the available devices.

    Attributes:
        tensor_parallelism: The degree of tensor parallelism (e.g., splitting
            weights of a single layer).
        expert_parallelism: The degree of expert parallelism for MoE models.
        sequence_parallelism: The degree of sequence parallelism (splitting
            activations along the sequence length dimension).
        data_parallelism: The degree of data parallelism (splitting the batch
            across devices).
    """
    tensor_parallelism: int = 1
    expert_parallelism: int = 1
    sequence_parallelism: int = 1
    data_parallelism: int = 1
    attention_data_parallelism: int = 1


class ShardingConfigManager:
    """Manages sharding configuration parsing and access from vLLM config.

    Usage:
        sharding_config = ShardingConfigManager.from_vllm_config(vllm_config)
        tp_size = sharding_config.tp_size

    During initialization, we set `vllm_config.sharding_config` to
    `ShardingConfigManager.from_vllm_config(vllm_config)`, so you can access
    `vllm_config.sharding_config.tp_size` directly.
    """

    def __init__(self,
                 sharding_strategy: ShardingStrategy,
                 device_indexes: Optional[List] = None):

        self.sharding_strategy: ShardingStrategy = sharding_strategy
        self.device_indexes: Optional[List[int]] = device_indexes
        self._total_devices: int = int(
            math.prod(asdict(sharding_strategy).values()))
        if device_indexes:
            assert self._total_devices == len(device_indexes)

    @classmethod
    def from_vllm_config(cls,
                         vllm_config: 'VllmConfig') -> 'ShardingConfigManager':

        sharding_strategy = vllm_config.additional_config.get(
            "sharding", {}).get("sharding_strategy", {})
        parallel_config = vllm_config.parallel_config
        tensor_parallelism = parallel_config.tensor_parallel_size
        data_parallelism = parallel_config.data_parallel_size
        expert_parallelism = sharding_strategy.get("expert_parallelism", 1)
        sequence_parallelism = sharding_strategy.get("sequence_parallelism", 1)
        device_indexes = sharding_strategy.get("device_indexes", None)

        enable_dp_attention = sharding_strategy.get("enable_dp_attention",
                                                    False)
        if enable_dp_attention:
            # Replicate attention layer when num_kv_heads < TP
            num_kv_heads = 1 if vllm_config.model_config.use_mla else vllm_config.model_config.get_total_num_kv_heads(
            )
            cache_dtype = vllm_config.cache_config.cache_dtype
            if cache_dtype == 'auto':
                cache_dtype = vllm_config.model_config.dtype
            kv_dtype = utils.get_jax_dtype_from_str_dtype(
                cache_dtype) or jnp.bfloat16
            packing = 4 // jnp.dtype(kv_dtype).itemsize
            # When num_kv_heads * 2 / packing < TP, tensor parallelism would
            # duplicate KV heads across devices, wasting kv cache memory.
            # Use attention DP instead to reduce per-device num_kv_heads and
            # eliminate this waste.
            num_kv_heads_per_device_in_kv_cache = (num_kv_heads * 2) / packing
            attn_dp = max(
                int(tensor_parallelism // num_kv_heads_per_device_in_kv_cache),
                1)
            tensor_parallelism = tensor_parallelism // attn_dp
        else:
            attn_dp = 1

        sharding_strategy = ShardingStrategy(
            tensor_parallelism=tensor_parallelism,
            data_parallelism=data_parallelism,
            expert_parallelism=expert_parallelism,
            sequence_parallelism=sequence_parallelism,
            attention_data_parallelism=attn_dp)

        # Must override here to avoid vLLM spinning up multiple DP engines.
        if vllm_config.parallel_config.data_parallel_size > 1:
            vllm_config.parallel_config.data_parallel_size = 1
            vllm_config.parallel_config.data_parallel_rank = 0
            vllm_config.parallel_config.data_parallel_size_local = 1

        cls.validate(vllm_config, sharding_strategy)
        return cls(sharding_strategy, device_indexes)

    @classmethod
    def validate(cls, vllm_config, sharding_strategy):
        total_dp_size = sharding_strategy.data_parallelism * sharding_strategy.attention_data_parallelism
        if total_dp_size > 1:
            if vllm_config.speculative_config is not None:
                raise ValueError(
                    f"Speculative decoding is not supported with data parallelism "
                    f"(DP size: {total_dp_size}). Please disable speculative decoding or "
                    f"set data parallelism to 1.")
            if vllm_config.lora_config is not None:
                raise ValueError(
                    f"LoRA is not supported with data parallelism "
                    f"(DP size: {total_dp_size}). Please disable LoRA or "
                    f"set data parallelism to 1.")
        if sharding_strategy.attention_data_parallelism > 1:
            if not envs.NEW_MODEL_DESIGN:
                raise ValueError(
                    "Must run Attention DP with NEW_MODEL_DESIGN enabled. Please set the "
                    "NEW_MODEL_DESIGN=True.")

    @property
    def total_dp_size(self) -> int:
        return self.sharding_strategy.data_parallelism * self.sharding_strategy.attention_data_parallelism

    @property
    def model_dp_size(self) -> int:
        return self.sharding_strategy.data_parallelism

    @property
    def attn_dp_size(self) -> int:
        return self.sharding_strategy.attention_data_parallelism

    @property
    def tp_size(self) -> int:
        return self.sharding_strategy.tensor_parallelism

    @property
    def expert_size(self) -> int:
        return self.sharding_strategy.expert_parallelism

    @property
    def sequence_size(self) -> int:
        return self.sharding_strategy.sequence_parallelism

    @property
    def total_devices(self) -> int:
        return self._total_devices

    def __str__(self):
        return (f"ShardingConfigManager(total_devices={self.total_devices}, "
                f"sharding_strategy={self.sharding_strategy}, "
                f"device_indexes={self.device_indexes})")


#TODO split this into block unique sharding config, i.e. attentionShardingConfig, MoEShardingConfig
@dataclass
class ShardingRulesConfig:
    """Holds detailed sharding configurations for individual tensors, namely logical rules.

    Each attribute in this class corresponds to a specific weight or activation
    tensor within a transformer model. The value of each attribute is a
    tuple of logical mesh axis names (e.g., 'dp', 'sp', 'tp'), which defines
    how the corresponding tensor's dimensions are partitioned across the device mesh.
    The dimension order in the attribute name (e.g., `btd` for batch, sequence,
    d_model) maps directly to the sharding tuple.

    TODO: update the mesh axis names to be clear and reduce confusion between prefill & generate
    """

    # Activation for attn input: (Batch * Sequence, Dim)
    activation_attention_td: tuple = (None, None)
    # Activation for attn out: (Batch * Sequence, Dim)
    activation_attention_out_td: tuple = (None, None)
    # Activation for q projection input: (Batch * Sequence, Dim)
    activation_q_td: tuple = (None, None)
    # Attention Out activation after projection: (Batch * Sequence, NumHeads, HeadDim)
    attn_o_tnh: tuple = (None, None, None)
    # Q vector: (Batch * Sequence, NumHeads, HeadDim)
    query_tnh: tuple = (None, None, None)
    # K/V vector: (Batch * Sequence, NumKVHeads, HeadDim)
    keyvalue_skh: tuple = (None, None, None)

    # Attention Q weight: (Dim, NumHeads, HeadDim)
    attn_q_weight_dnh: tuple = (None, None, None)
    # Attention K weight: (Dim, NumKVHeads, HeadDim)
    attn_k_weight_dkh: tuple = (None, None, None)
    # Attention V weight: (Dim, NumKVHeads, HeadDim)
    attn_v_weight_dkh: tuple = (None, None, None)
    # Attention Out weight: (NumHeads, HeadDim, Dim)
    attn_o_weight_nhd: tuple = (None, None, None)

    # Activation for ffw input: (Batch * Sequence, Dim)
    activation_ffw_td: tuple = (None, None)

    # Activation for ffw input: (Batch * Sequence, Expert, Dim)
    activation_ffw_ted: tuple = (None, None, None)

    # FFW hidden activation: (Batch * Sequence, FfwDim)
    ffw_hidden_tf: tuple = (None, None)

    # FFW up/gate weight: (Dim, FfwDim)
    ffw_weight_df: tuple = (None, None)
    # FFW down weight: (FfwDim, Dim)
    ffw_weight_fd: tuple = (None, None)
    # MoE gate/up weights: (NumExperts, Dim, FfwDim)
    moe_weights_edf: tuple = (None, None, None)
    # MoE down weights: (NumExperts, FfwDim, Dim)
    moe_weights_efd: tuple = (None, None, None)
    # MoE router weights: (Dim, NumExperts)
    moe_router_de: tuple = (None, None)
    # MoE router bias weights: (NumExperts,)
    moe_router_bias_e: tuple = (None, )

    # Embedding weight: (VocabSize, Dim)
    emb_weight_vd: tuple = (None, None)
    # Activation between layers: (Batch * Sequence, Dim)
    activation_td: tuple = (None, None)
    # Final activation before logits: (Batch * Sequence, Dim)
    prelogit_td: tuple = (None, None)
    # Logit activation: (Batch * Sequence, VocabSize)
    logits_tv: tuple = (None, None)
    # RMS norm scale weight: (Dim,)
    norm_scale: tuple = (None)
    # Vocab projection weight (tied embeddings): (Dim, VocabSize)
    vocab_vd: tuple = (None, None)
    vocab_dv: tuple = (None, None)


class ShardingConfig:
    """Container for operation-specific sharding configurations.

    This class holds two separate `ShardingRulesConfig` objects, one for the
    'prefill' phase and one for the 'generate' (or decode) phase of model
    execution. This allows tailoring sharding strategies to the different
    computational patterns of each phase.

    Example Sharding Strategy and Configuration:

    Sharding Strategy defines the high-level parallelism dimensions.
    For a device mesh like `Mesh((2, 4, 4, 4), ('data', 'seq', 'expert', 'tensor'))` on 128 devices:
    - data: Data Parallelism (2-way)
    - seq: Sequence Parallelism (4-way)
    - expert: Expert Parallelism (4-way)
    - tensor: Tensor Parallelism (4-way)

    ShardingConfig then maps tensor dimensions to these logical mesh axes.
    For example, a tensor with shape (Batch, Sequence, Dimension) could be sharded
    differently for prefill and decode/generate operations:

    - Prefill (long sequences, small batch):
    Sharding sequence dim on the 'sp' axis is often efficient.
    `prefill_rules.activation_attention_btd = (None, 'seq', 'tensor')`

    - Generate (short sequences, large batch):
    Sharding batch dim on the 'dp' axis is often efficient.
    `generate_rules.activation_attention_btd = ('data', None, 'tensor')`
    """

    def __init__(self,
                 prefill_rules=None,
                 generate_rules=None,
                 default_rules_cls=ShardingRulesConfig):
        """Initializes the ShardingConfig.

        Args:
            prefill_rules: An `ShardingRulesConfig` for the prefill phase.
                If None, a default config is created.
            generate_rules: An `ShardingRulesConfig` for the generate phase.
                If None, a default config is created.
            default_rules_cls: The default sharding rules (class) to use.
        """
        # Use a factory pattern to avoid mutable default arguments
        self.default_rules_cls = default_rules_cls
        self.prefill_rules = prefill_rules if prefill_rules is not None else default_rules_cls(
        )
        self.generate_rules = generate_rules if generate_rules is not None else default_rules_cls(
        )


def build_mesh(devices, strategy: dict[str, int]) -> Mesh:
    """Constructs a JAX device mesh from a sharding strategy.

    This method creates a logical grid of devices based on the parallelism
    degrees defined in the strategy. The logical axis names ('dp', 'ep',
    'sp', 'tp') are used to map tensor dimensions to the physical device grid.

    Args:
        strategy: A dictionary from upper level config.

    Returns:
        A JAX `Mesh` object.
    """

    axis_order = {
        "data": strategy.get("data_parallelism", 1),
        "expert": strategy.get("expert_parallelism", 1),
        "seq": strategy.get("sequence_parallelism", 1),
        "model": strategy.get("tensor_parallelism", 1),
    }
    # TODO: add logic to infer axis when the degree is -1
    mesh_axis_names = []
    mesh_shape = []
    for axis, dim in axis_order.items():
        mesh_axis_names.append(axis)
        mesh_shape.append(dim)

    if not mesh_shape:
        mesh_shape = [1]
        mesh_axis_names = [
            'data'
        ]  # default to data parallelism if no other strategy is specified

    devices = np.asarray(devices).reshape(mesh_shape)
    return Mesh(devices, axis_names=tuple(mesh_axis_names))


class Sharding:
    """Generates and manages sharding configurations based on a high-level strategy.

    This class populates a `ShardingConfig` with detailed tensor sharding
    rules for both prefill and generation phases. It also allows for runtime
    overrides of these rules.

    Attributes:
        sharding_cfg: The generated `ShardingConfig` with detailed rules.
    """

    def __init__(self,
                 prefill_rules: dict | None = None,
                 generate_rules: dict | None = None,
                 default_rules_cls=ShardingRulesConfig,
                 vllm_config: 'VllmConfig' = None):
        """Initializes the Sharding manager.

        Args:
            prefill_rules: A dictionary of overrides for the prefill
                sharding config. Keys are attribute names in `ShardingRulesConfig`,
                and values are the new sharding tuples.
            generate_rules: A dictionary of overrides for the generate
                sharding config.
        """
        self.vllm_config = vllm_config
        self.default_rules_cls = default_rules_cls
        self.sharding_cfg = self.make_sharding_config(
            default_rules_cls=default_rules_cls,
            prefill_overrides=prefill_rules,
            generate_overrides=generate_rules)

    def _get_overrides(self, sharding_phase: str):
        """Return the overrides from the vLLM config for the given sharding phase."""
        overrides = {}
        try:
            overrides = self.vllm_config.additional_config["sharding"][
                "logical_rules"]["all"]
        except KeyError:
            pass

        try:
            additional_overrides = self.vllm_config.additional_config[
                "sharding"]["logical_rules"][f"{sharding_phase}"]
            overrides.update(additional_overrides)
        except KeyError:
            pass
        return overrides

    def __str__(self):
        """Succinct representation of relevant Sharding settings and overrides."""
        output_str = f"  Using {self.default_rules_cls.__name__} logical rules.\n"
        output_str += f"  {self.__class__.__name__:} overrides:\n"
        output_str += f"    prefill logical_rule overrides:\n    {json.dumps(self._get_overrides('prefill'), indent=4, default=str)}\n\n"
        output_str += f"    generate logical_rule overrides:\n    {json.dumps(self._get_overrides('generate'), indent=4, default=str)}\n\n"
        return output_str

    def validate_sharding_strategy(self, ):
        """Validates if the sharding strategy is compatible with the environment.

        This method is a placeholder now, and will check if the product of parallelism degrees
        matches the number of available devices.
        """
        #TODO: check num_devices % parallelism == 0
        #TODO: check num_devices == multiply(parallelism(with inferred))
        return

    def get_sharding_cfg(self) -> ShardingConfig:
        """Returns the generated sharding configuration."""
        return self.sharding_cfg

    def _apply_overrides(self, config_obj: ShardingRulesConfig,
                         overrides: dict | None):
        """Applies runtime overrides to a sharding configuration object.

        Args:
            config_obj: The sharding configuration object (e.g., prefill_rules)
                to be updated.
            overrides: A dictionary where keys are attribute names of the config
                object and values are the new sharding tuples.

        Raises:
            AttributeError: If a key in the overrides dictionary is not a valid
                attribute of the configuration object.
        """
        for key, value in overrides.items():
            if hasattr(config_obj, key):
                setattr(config_obj, key, value)
            else:
                # Raise an error for invalid keys to prevent silent failures
                raise AttributeError(
                    f"'{key}' is not a valid attribute of {type(config_obj).__name__}"
                )

    def _make_default_sharding_config(self, prefill_rules, generate_rules):

        # Populate Prefill Config
        # During prefill, sequence length is long, so we shard along the sequence axis.
        prefill_rules.activation_attention_td = (ShardingAxisName.ATTN_DATA,
                                                 ShardingAxisName.ATTN_TENSOR)
        prefill_rules.activation_attention_out_td = (
            ShardingAxisName.ATTN_DATA, ShardingAxisName.ATTN_TENSOR)
        prefill_rules.activation_q_td = (ShardingAxisName.ATTN_DATA,
                                         ShardingAxisName.ATTN_TENSOR)
        #TODO: the default qkv and kvcache is sharded on head dim
        # We may change it after we finalize the KVCache design
        prefill_rules.attn_o_tnh = (ShardingAxisName.ATTN_DATA,
                                    ShardingAxisName.ATTN_HEAD, None)
        prefill_rules.query_tnh = (ShardingAxisName.ATTN_DATA,
                                   ShardingAxisName.ATTN_HEAD, None)
        prefill_rules.keyvalue_skh = (ShardingAxisName.ATTN_DATA,
                                      ShardingAxisName.ATTN_HEAD, None)

        # Populate Generate (Decode) Config
        # During decode, batch size is the large dimension, so we shard along the batch axis.
        generate_rules.activation_attention_td = (ShardingAxisName.ATTN_DATA,
                                                  ShardingAxisName.ATTN_TENSOR)
        generate_rules.activation_attention_out_td = (
            ShardingAxisName.MLP_DATA, ShardingAxisName.ATTN_TENSOR)
        generate_rules.activation_q_td = (ShardingAxisName.ATTN_DATA,
                                          ShardingAxisName.ATTN_TENSOR)
        #TODO: the default qkv and kvcache is sharded on head dim
        # We may change it after we finalize the KVCache design
        generate_rules.attn_o_tnh = (ShardingAxisName.ATTN_DATA,
                                     ShardingAxisName.ATTN_HEAD, None)
        generate_rules.query_tnh = (ShardingAxisName.ATTN_DATA,
                                    ShardingAxisName.ATTN_HEAD, None)
        generate_rules.keyvalue_skh = (ShardingAxisName.ATTN_DATA,
                                       ShardingAxisName.ATTN_HEAD, None)
        generate_rules.attn_q_weight_dnh = (None, ShardingAxisName.ATTN_HEAD,
                                            ShardingAxisName.ATTN_TENSOR)
        generate_rules.attn_k_weight_dkh = (None, ShardingAxisName.ATTN_HEAD,
                                            ShardingAxisName.ATTN_TENSOR)
        generate_rules.attn_v_weight_dkh = (None, ShardingAxisName.ATTN_HEAD,
                                            ShardingAxisName.ATTN_TENSOR)
        generate_rules.attn_o_weight_nhd = (ShardingAxisName.ATTN_HEAD, None,
                                            ShardingAxisName.ATTN_TENSOR)
        generate_rules.activation_ffw_td = (ShardingAxisName.MLP_DATA, None)
        generate_rules.activation_ffw_ted = (ShardingAxisName.MLP_DATA,
                                             ShardingAxisName.EXPERT, None)
        generate_rules.ffw_hidden_tf = (ShardingAxisName.MLP_DATA,
                                        ShardingAxisName.MLP_TENSOR)
        # FFW weights are typically sharded along the hidden dimension (F).
        generate_rules.ffw_weight_df = (None, ShardingAxisName.MLP_TENSOR)
        generate_rules.ffw_weight_fd = (ShardingAxisName.MLP_TENSOR, None)
        # MoE weights are sharded along the expert axis and the hidden dimension.
        generate_rules.moe_weights_edf = (ShardingAxisName.EXPERT, None,
                                          ShardingAxisName.MOE_TENSOR)
        generate_rules.moe_weights_efd = (ShardingAxisName.EXPERT,
                                          ShardingAxisName.MOE_TENSOR, None)
        generate_rules.moe_router_de = (None, ShardingAxisName.EXPERT)

        # Embedding weight: (VocabSize, Dim)
        generate_rules.emb_weight_vd = (ShardingAxisName.MLP_TENSOR, None)
        generate_rules.activation_td = (ShardingAxisName.MLP_DATA,
                                        ShardingAxisName.ATTN_TENSOR)
        generate_rules.prelogit_td = (ShardingAxisName.MLP_DATA,
                                      ShardingAxisName.MLP_TENSOR)
        generate_rules.logits_tv = (ShardingAxisName.MLP_DATA,
                                    ShardingAxisName.MLP_TENSOR)
        generate_rules.vocab_vd = (ShardingAxisName.VOCAB, None)
        generate_rules.vocab_dv = (None, ShardingAxisName.VOCAB)

    def make_sharding_config(
            self,
            default_rules_cls: ShardingRulesConfig,
            prefill_overrides: dict | None = None,
            generate_overrides: dict | None = None) -> ShardingConfig:
        """Creates the detailed `ShardingConfig` with specific partitioning rules
        and applies any runtime overrides.

        This method populates the `prefill_rules` and
        `generate_rules` with hardcoded sharding rules that are generally
        effective for transformer models, and then updates them with any provided
        overrides.

        Args:
            prefill_overrides: A dictionary with attribute names and their new values
                for the prefill sharding configuration.
            generate_overrides: A dictionary with attribute names and their new values
                for the generate sharding configuration.

        Returns:
            The populated and overridden `ShardingConfig` object.
        """
        #TODO: organize into update_prefill() and update_decode for each axis
        #TODO: verify the sharding axes
        sharding_cfg = ShardingConfig(default_rules_cls=default_rules_cls)
        prefill_rules = sharding_cfg.prefill_rules
        generate_rules = sharding_cfg.generate_rules

        # Extract the overrides from the vllm_config if they are not provided programatically.
        if prefill_overrides is None:
            prefill_overrides = self._get_overrides("prefill")
        if generate_overrides is None:
            generate_overrides = self._get_overrides("generate")

        # Apply default sharding configs
        self._make_default_sharding_config(prefill_rules, generate_rules)

        # Apply overriding the runtime sharding rules
        self._apply_overrides(prefill_rules, prefill_overrides)
        self._apply_overrides(generate_rules, generate_overrides)

        return sharding_cfg

    #TODO: Add __repr__


class ShardingInfo:
    #TODO a sharding info class for visualizing & debugging the sharding performance
    # Will implement it for the next version
    pass



from typing import Union

import jax
import jax.numpy as jnp
import torch
import torch.nn.functional as F
from jax.experimental.layout import Format, Layout
from jax.sharding import Mesh, NamedSharding
from jax.sharding import PartitionSpec as P
from torch.nn.parameter import Parameter
from torchax.interop import call_jax, torch_view
from torchax.ops.mappings import t2j
from vllm.logger import init_logger
from vllm.model_executor.layers.fused_moe import FusedMoE, FusedMoEConfig
from vllm.model_executor.layers.quantization.compressed_tensors.compressed_tensors import \
    CompressedTensorsConfig
from vllm.model_executor.layers.quantization.compressed_tensors.compressed_tensors_moe import \
    CompressedTensorsW8A8Fp8MoEMethod
from vllm.model_executor.layers.quantization.compressed_tensors.schemes.compressed_tensors_wNa16 import (  # noqa
    WNA16_SUPPORTED_BITS, WNA16_SUPPORTED_TYPES_MAP)

from tpu_inference.layers.vllm.quantization.common import JaxCommonConfig

logger = init_logger(__name__)


class VllmCompressedTensorsW8A8Fp8MoEMethod(CompressedTensorsW8A8Fp8MoEMethod,
                                            JaxCommonConfig):

    def __init__(self, quant_config: "CompressedTensorsConfig",
                 moe: FusedMoEConfig, mesh: Mesh):
        super().__init__(quant_config, moe)
        self.mesh = mesh
        self.quant_config = quant_config

        # disable GPU paths
        self.use_marlin = False
        self.rocm_aiter_moe_enabled = False  # is_rocm_aiter_moe_enabled()
        self.is_fp8_w8a8_sm100 = False
        self.use_cutlass = False
        self.disable_expert_map = False

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        assert isinstance(layer, FusedMoE)

        intermediate_size = layer.w13_weight.shape[1] // 2
        w1_weight = layer.w13_weight[:, :intermediate_size]
        w3_weight = layer.w13_weight[:, intermediate_size:]
        w1_weight_scale = layer.w13_weight_scale[:, :intermediate_size]
        w3_weight_scale = layer.w13_weight_scale[:, intermediate_size:]

        w2_weight = t2j(layer.w2_weight, use_dlpack=False)
        w2_weight_scale = t2j(layer.w2_weight_scale.to(torch.bfloat16),
                              use_dlpack=False)
        w1_weight = t2j(w1_weight, use_dlpack=False)
        w1_weight_scale = t2j(w1_weight_scale.to(torch.bfloat16),
                              use_dlpack=False)
        w3_weight = t2j(w3_weight, use_dlpack=False)
        w3_weight_scale = t2j(w3_weight_scale.to(torch.bfloat16),
                              use_dlpack=False)

        if layer.use_ep:
            format = Format(Layout((0, 1, 2)),
                            NamedSharding(self.mesh, P("model", None, None)))
            w1_weight = jax.device_put(w1_weight, format)
            w1_weight_scale = jax.device_put(w1_weight_scale, format)
            w3_weight = jax.device_put(w3_weight, format)
            w3_weight_scale = jax.device_put(w3_weight_scale, format)
            w2_weight = jax.device_put(w2_weight, format)
            w2_weight_scale = jax.device_put(w2_weight_scale, format)
        else:
            assert intermediate_size == w2_weight.shape[-1]
            n_shards = self.mesh.shape["model"]
            assert intermediate_size % n_shards == 0

            # TODO: enable this if using fused weights
            # output_sizes = [intermediate_size, intermediate_size]
            # w13_weight = reorder_concatenated_tensor_for_sharding(
            #    w13_weight, output_sizes, n_shards, dim=1
            # )

            w13_format = Format(
                Layout((0, 1, 2)),
                NamedSharding(self.mesh, P(None, "model", None)))
            w1_weight = jax.device_put(w1_weight, w13_format)
            w1_weight_scale = jax.device_put(w1_weight_scale, w13_format)
            w3_weight = jax.device_put(w3_weight, w13_format)
            w3_weight_scale = jax.device_put(w3_weight_scale, w13_format)
            w2_weight = jax.device_put(
                w2_weight,
                Format(Layout((0, 1, 2)),
                       NamedSharding(self.mesh, P(None, None, "model"))),
            )
            w2_weight_scale = jax.device_put(
                w2_weight_scale,
                Format(Layout((0, 1, 2)), NamedSharding(self.mesh, P())),
            )  # replicate

        w1_weight = Parameter(torch_view(w1_weight), requires_grad=False)
        w1_weight_scale = Parameter(torch_view(w1_weight_scale),
                                    requires_grad=False)
        w2_weight = Parameter(torch_view(w2_weight), requires_grad=False)
        w2_weight_scale = Parameter(torch_view(w2_weight_scale),
                                    requires_grad=False)
        w3_weight = Parameter(torch_view(w3_weight), requires_grad=False)
        w3_weight_scale = Parameter(torch_view(w3_weight_scale),
                                    requires_grad=False)

        # TODO dont reuse variable
        layer.w13_weight = w1_weight
        layer.w13_weight_scale = w1_weight_scale
        layer.w2_weight = w2_weight
        layer.w2_weight_scale = w2_weight_scale
        layer.w3_weight = w3_weight
        layer.w3_weight_scale = w3_weight_scale

    def apply(
        self,
        layer: torch.nn.Module,
        x: torch.Tensor,
        router_logits: torch.Tensor,
    ) -> Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        assert isinstance(layer, FusedMoE)
        if layer.activation != "silu":
            raise NotImplementedError(
                "Only silu is supported for activation function.")
        if layer.scoring_func != "softmax":
            raise NotImplementedError(
                "Only softmax is supported for scoring_func")

        # import sys
        # sys.stdin = open(0)
        # breakpoint()

        # TODO: Use MoE kernel when it supports fp8

        seqlen = x.shape[0]

        expert_weights = F.softmax(router_logits, dim=-1)
        expert_weights, expert_indices = torch.topk(expert_weights,
                                                    layer.top_k,
                                                    dim=-1)
        if layer.renormalize:
            expert_weights /= expert_weights.sum(dim=-1, keepdim=True)

        # cond ffn
        # e = total num of exp = 160
        # t = seqlen
        # o = config.imtermediate size
        # i = config.dim
        #torch.einsum("ti, eoi -> teo", x, layer.w13_weight) * self.w13_weight_scale)
        ux1 = call_jax(jax.lax.dot,
                       x,
                       layer.w13_weight,
                       dimension_numbers=(((1, ), (2, )), ((), ())),
                       preferred_element_type=jnp.bfloat16.dtype)
        x1 = F.silu(ux1 * layer.w13_weight_scale.squeeze(2))

        #x3 = torch.einsum("ti, eoi -> teo", x, layer.w3_weight) * self.w3_weight_scale
        x3 = call_jax(jax.lax.dot,
                      x,
                      layer.w3_weight,
                      dimension_numbers=(((1, ), (2, )), ((), ())),
                      preferred_element_type=jnp.bfloat16.dtype
                      ) * layer.w3_weight_scale.squeeze(2)

        #expert_outs = torch.einsum("teo, eio -> tei", (x1 * x3), self.w2_weight) * self.w2_weight_scale
        expert_outs = call_jax(
            jax.lax.dot,
            x1 * x3,
            layer.w2_weight,
            dimension_numbers=(((2, ), (2, )), ((1, ), (0, ))),
            preferred_element_type=jnp.bfloat16.dtype).transpose(
                0, 1) * layer.w2_weight_scale.squeeze(2)

        seq_indexes = torch.arange(seqlen, device='jax').unsqueeze(1)
        expert_outs = expert_outs[seq_indexes, expert_indices]

        # out = torch.einsum("tai,ta -> ti", expert_outs, expert_weights)
        out = call_jax(jax.lax.dot,
                       expert_outs,
                       expert_weights,
                       dimension_numbers=(((1, ), (1, )), ((0, ), (0, ))),
                       preferred_element_type=jnp.bfloat16.dtype)

        return out



import math
from typing import Any, Dict

import jax
import jax.numpy as jnp


def apply_rope(
    # (seq_len, num_heads, head_dim)
    inputs: jax.Array,
    # (3, seq_len) for M-RoPE, otherwise (seq_len,)
    positions: jax.Array,
    head_dim: int,
    rope_theta: float = 10000,
    rope_scaling: Dict[str, Any] = None,
    rope_input_ordering: str = "split",
) -> jax.Array:
    """
    Applies Rotary Positional Embedding using the sine and cosine strategy.

    This implementation assumes the input tensor has a shape that might include
    padding on the last dimension (head_dim).
    RoPE is applied only to the first `head_dim` features, and the result is
    padded back to the original dimension if necessary.
    If rope_input_ordering is "split", then the input pairs for rotation are taken one from the
    first and one from the second half of the head_dim. If it is "interleaved" then
    adjacent values are used as inputs for rotation.
    """

    # M-RoPE support for Qwen2.5-VL
    if positions.ndim == 2 and positions.shape[0] == 3:
        mrope_section = rope_scaling.get("mrope_section",
                                         None) if rope_scaling else None
        # NOTE: We assume mrope_section is always available
        # as Qwen2.5-VL is the only model using mrope
        assert mrope_section is not None

        split_indices = [mrope_section[0], mrope_section[0] + mrope_section[1]]

        # Indices for the features to be rotated (first half of head_dim)
        all_freq_indices = jnp.arange(head_dim // 2)

        # Split the indices according to mrope_section. This is valid because split_indices are static.
        freq_indices_split = jnp.split(all_freq_indices, split_indices)
        # freq_indices_split is a list of 3 JAX arrays.

        cos_list = []
        sin_list = []

        for i in range(3):  # For each of the 3 position dimensions
            current_indices = freq_indices_split[i]

            if current_indices.size == 0:
                # This section is empty, skip.
                continue

            # inv_freq shape: (mrope_section[i],)
            inv_freq = 1.0 / (rope_theta**(current_indices * 2.0 / head_dim))

            # positions[i]: (seq_len,)
            # freqs shape: (seq_len, mrope_section[i])
            freqs = jnp.outer(positions[i], inv_freq)

            cos_list.append(jnp.cos(freqs))
            sin_list.append(jnp.sin(freqs))

        # Concatenate along the feature dimension
        # cos, sin shape: (seq_len, head_dim//2)
        cos = jnp.concatenate(cos_list, axis=1)
        sin = jnp.concatenate(sin_list, axis=1)

        # Add num_heads dimension for broadcasting
        cos = cos[:, jnp.newaxis, :]  # Shape: (seq_len, 1, head_dim//2)
        sin = sin[:, jnp.newaxis, :]  # Shape: (seq_len, 1, head_dim//2)

        # Apply rotation
        inputs_real = inputs[..., :head_dim // 2]
        inputs_imag = inputs[..., head_dim // 2:head_dim]

        outputs_real = inputs_real * cos - inputs_imag * sin
        outputs_imag = inputs_real * sin + inputs_imag * cos

        out = jnp.concatenate([outputs_real, outputs_imag], axis=-1)

    # Standard RoPE
    else:
        # Calculate inverse frequencies (timescale)
        fraction = 2 * jnp.arange(0, head_dim // 2) / head_dim
        timescale = 1.0 / (rope_theta**fraction)

        # Apply scaling if provided
        if rope_scaling:
            timescale = apply_rope_scaling(timescale, rope_scaling)

        # Prepare for rotation by calculating sin and cos values
        # `sinusoid_inp` gets shape (batch * seq_len, head_dim/2)
        sinusoid_inp = positions[..., jnp.newaxis] * timescale[jnp.newaxis, :]

        # Broadcast over the 'heads' dimension, assuming shape (batch*seq, heads, head_dim)
        sinusoid_inp = sinusoid_inp[:, jnp.newaxis, ...]
        sin = jnp.sin(sinusoid_inp)
        cos = jnp.cos(sinusoid_inp)

        if rope_input_ordering == "interleaved":
            # Reshape to group adjacent features for rotation, matching new_apply_rope
            rotary_inputs = inputs[
                ..., :head_dim]  # Take just the non-padded amount.
            reshaped_inputs = rotary_inputs.reshape(*rotary_inputs.shape[:-1],
                                                    -1, 2)

            # Apply the rotation
            first_half = reshaped_inputs[..., 0]
            second_half = reshaped_inputs[..., 1]
        else:
            first_half = inputs[..., :head_dim // 2]
            second_half = inputs[..., head_dim // 2:head_dim]

        first_part = first_half * cos - second_half * sin
        second_part = second_half * cos + first_half * sin

        # Combine the rotated parts and reshape back
        if rope_input_ordering == "interleaved":
            out_stacked = jnp.stack([first_part, second_part], axis=-1)
            out = out_stacked.reshape(rotary_inputs.shape)
        else:
            out = jnp.concatenate([first_part, second_part], axis=-1)

    # If the original input was padded, pad the output with zeros to match.
    padded_head_dim = inputs.shape[-1]
    if padded_head_dim > head_dim:
        pad_width = padded_head_dim - head_dim
        pad_config = [(0, 0)] * (out.ndim - 1) + [(0, pad_width)]
        out = jnp.pad(out, pad_config)

    return out.astype(inputs.dtype)


def apply_longrope(
    inputs: jax.Array,
    positions: jax.Array,
    head_dim: int,
    rope_scaling: Dict[str, Any],
    original_max_position_embeddings: int,
    max_position_embeddings: int,
    rope_theta: float = 10000,
) -> jax.Array:
    # LongRoPE implementation specific to Phi-3
    # Implementation based on https://github.com/huggingface/transformers/blob/main/src/transformers/models/phi3/modeling_phi3.py#L197-L235

    scale = max_position_embeddings / original_max_position_embeddings
    if scale <= 1.0:
        mscale = 1.0
    else:
        mscale = jnp.sqrt(1 + (jnp.log(scale) /
                               jnp.log(original_max_position_embeddings)))

    seq_len = inputs.shape[0]
    if seq_len > original_max_position_embeddings:
        long_factor = jnp.array(rope_scaling.get("long_factor"))
        timescale = 1.0 / (long_factor * (rope_theta**(
            (2 * jnp.arange(0, head_dim // 2)) / head_dim)))
    else:
        short_factor = jnp.array(rope_scaling.get("short_factor"))
        timescale = 1.0 / (short_factor * (rope_theta**(
            (2 * jnp.arange(0, head_dim // 2)) / head_dim)))

    # Calculate RoPE positions
    sinusoid_inp = positions[..., jnp.newaxis] * timescale[jnp.newaxis, :]
    sinusoid_inp = sinusoid_inp[:, jnp.newaxis, ...]
    sin = jnp.sin(sinusoid_inp) * mscale
    cos = jnp.cos(sinusoid_inp) * mscale

    # Padding logic
    padded_head_dim = inputs.shape[-1]

    # Apply RoPE mechanism
    first_half = inputs[..., :head_dim // 2]
    second_half = inputs[..., head_dim // 2:head_dim]

    first_part = first_half * cos - second_half * sin
    second_part = second_half * cos + first_half * sin
    out = jnp.concatenate([first_part, second_part], axis=-1)

    if padded_head_dim > head_dim:
        out = jnp.pad(out, ((0, 0), (0, 0), (0, padded_head_dim - head_dim)))

    return out.astype(inputs.dtype)


def apply_rope_scaling(freqs: jax.Array, rope_scaling: Dict[str,
                                                            Any]) -> jax.Array:
    # Values obtained from grid search
    scale_factor = rope_scaling.get("scale_factor", 8.0)
    low_freq_factor = rope_scaling.get("low_freq_factor", 1.0)
    high_freq_factor = rope_scaling.get("high_freq_factor", 4.0)
    old_context_len = rope_scaling.get("original_max_position_embeddings",
                                       8192)

    low_freq_wavelen = old_context_len / low_freq_factor
    high_freq_wavelen = old_context_len / high_freq_factor

    wavelen = 2 * math.pi / freqs
    smooth = (old_context_len / wavelen -
              low_freq_factor) / (high_freq_factor - low_freq_factor)

    high_freqs = jnp.where(wavelen < high_freq_wavelen, freqs, 0)
    low_freqs = jnp.where(wavelen > low_freq_wavelen, freqs / scale_factor, 0)
    mid_freqs = jnp.where(
        (wavelen >= high_freq_wavelen) & (wavelen <= low_freq_wavelen),
        (1 - smooth) * freqs / scale_factor + smooth * freqs,
        0,
    )
    new_freqs = high_freqs + low_freqs + mid_freqs
    return new_freqs



# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Optional, Tuple, Union, cast

import jax.numpy as jnp
import torch
import vllm.envs as vllm_envs
from tpu_info import device
from vllm.inputs import ProcessorInputs, PromptType
from vllm.platforms.interface import Platform, PlatformEnum

from tpu_inference import envs
from tpu_inference.layers.common.sharding import ShardingConfigManager
from tpu_inference.logger import init_logger
from tpu_inference.utils import to_jax_dtype, to_torch_dtype

if TYPE_CHECKING:
    from vllm.attention.backends.registry import AttentionBackendEnum
    from vllm.config import BlockSize, ModelConfig, VllmConfig
    from vllm.pooling_params import PoolingParams
    from vllm.sampling_params import SamplingParams, SamplingType
else:
    BlockSize = None
    ModelConfig = None
    VllmConfig = None
    PoolingParams = None
    AttentionBackendEnum = None
    SamplingParams = None
    SamplingType = None

logger = init_logger(__name__)


class TpuPlatform(Platform):
    _enum = PlatformEnum.TPU
    device_name: str = "tpu"
    device_type: str = "tpu"
    dispatch_key: str = "XLA"
    ray_device_key: str = "TPU"
    device_control_env_var: str = "TPU_VISIBLE_CHIPS"
    simple_compile_backend: str = "openxla"

    supported_quantization: list[str] = [
        "tpu_int8", "compressed-tensors", "awq", "fp8", "mxfp4"
    ]

    additional_env_vars: list[str] = [
        "PHASED_PROFILING_DIR", "TPU_CHIPS_PER_HOST_BOUNDS", "TPU_HOST_BOUNDS",
        "TPU_MULTIHOST_BACKEND", "VLLM_MLA_DISABLE", "TPU_BACKEND_TYPE",
        "NEW_MODEL_DESIGN"
    ]

    @classmethod
    def get_attn_backend_cls(cls, selected_backend: "AttentionBackendEnum",
                             head_size: int, dtype: jnp.dtype,
                             kv_cache_dtype: Optional[str], block_size: int,
                             use_mla: bool, has_sink: bool, use_sparse: bool,
                             use_mm_prefix: bool, attn_type: Any) -> str:
        from vllm.attention.backends.registry import AttentionBackendEnum
        if selected_backend != AttentionBackendEnum.PALLAS:
            logger.info("Cannot use %s backend on TPU.", selected_backend)

        logger.info("Using Pallas V1 backend.")
        return "tpu_inference.layers.vllm.attention.PallasAttentionBackend"

    @classmethod
    def get_device_name(cls, device_id: int = 0) -> str:
        try:
            if vllm_envs.VLLM_TPU_USING_PATHWAYS:
                # Causes mutliprocess accessing IFRT when calling jax.devices()
                return "TPU v6 lite"
            else:
                chip_type, _ = device.get_local_chips()
                return f"TPU {chip_type.name}"
        except Exception as e:
            logger.warning(f"Error getting device name: {e}")
            return 'TPU'

    @classmethod
    def fp8_dtype(cls) -> torch.dtype:
        if cls.get_device_name().lower() == "tpu v6e":
            logger.info(
                "Automatically using fp8_e5m2 for FP8 KV cache on TPU v6e.")
            return torch.float8_e5m2
        return torch.float8_e4m3fn

    @classmethod
    def get_device_total_memory(cls, device_id: int = 0) -> int:
        raise NotImplementedError

    @classmethod
    def is_async_output_supported(cls, enforce_eager: Optional[bool]) -> bool:
        return False

    @classmethod
    def get_punica_wrapper(cls) -> str:
        return "tpu_inference.lora.torch_punica_tpu.PunicaWrapperTPU"

    @classmethod
    def get_infinity_values(cls, dtype: jnp.dtype) -> Tuple[float, float]:
        return jnp.finfo(dtype).min, jnp.finfo(dtype).max

    @classmethod
    def can_update_inplace(cls):
        return False

    @classmethod
    def get_lora_vocab_padding_size(cls) -> int:
        return 1

    @classmethod
    def inference_mode(cls):
        return True

    @classmethod
    def _initialize_sharding_config(cls, vllm_config: VllmConfig) -> None:

        sharding_config = ShardingConfigManager.from_vllm_config(vllm_config)
        vllm_config.sharding_config = sharding_config
        logger.info(f"Initialized sharding configuration: {sharding_config}")

    @classmethod
    def check_and_update_config(cls, vllm_config: VllmConfig) -> None:

        if vllm_envs.VLLM_TPU_USING_PATHWAYS:
            assert not vllm_envs.VLLM_ENABLE_V1_MULTIPROCESSING, (
                "VLLM_ENABLE_V1_MULTIPROCESSING must be 0 when using Pathways(JAX_PLATFORMS=proxy)"
            )
        cls._initialize_sharding_config(vllm_config)

        from vllm.config import CompilationMode

        cache_config = vllm_config.cache_config
        # For v0, the default block size is 16.
        if cache_config and cache_config.block_size is None:
            cache_config.block_size = cast(BlockSize, 16)

        compilation_config = vllm_config.compilation_config

        # TPU only supports DYNAMO_TRACE_ONCE compilation level
        # NOTE(xiang): the compilation_config is not used by jax.
        if compilation_config.mode != CompilationMode.DYNAMO_TRACE_ONCE:
            compilation_config.mode = CompilationMode.DYNAMO_TRACE_ONCE

        if compilation_config.backend == "":
            compilation_config.backend = "openxla"

        # If we use vLLM's model implementation in PyTorch, we should set it with torch version of the dtype.
        impl = envs.MODEL_IMPL_TYPE

        # NOTE(xiang): convert dtype to jnp.dtype
        # NOTE(wenlong): skip this logic for mm model preprocessing
        # For mm model preprocessors, it may need the output dtype to be torch.
        # In order to avoid a PR to vLLM, we postpone the dtype checking during
        # tpu_worker initialization
        if not vllm_config.scheduler_config.is_multimodal_model or impl == "vllm":
            model_dtype = vllm_config.model_config.dtype
            try:
                dtype = to_jax_dtype(model_dtype)
            except ValueError:
                logger.warning(f"{model_dtype=} is not supported. "
                               "Falling back to jnp.bfloat16")
                dtype = jnp.bfloat16
            if impl == "vllm":
                dtype = to_torch_dtype(dtype)
            vllm_config.model_config.dtype = dtype

        # TODO(cuiq): remove this dependency.
        from vllm.v1.attention.backends.pallas import PallasAttentionBackend
        cache_config.block_size = PallasAttentionBackend.get_page_size(
            vllm_config)  # type: ignore[assignment]
        min_page_size = PallasAttentionBackend.get_min_page_size(vllm_config)
        if min_page_size > cache_config.block_size:
            logger.warning(
                "Increase the page size from %s to %s to make sure there's"
                "no SMEM OOM",
                cache_config.block_size,
                min_page_size,
            )
            cache_config.block_size = min_page_size  # type: ignore[assignment]

        parallel_config = vllm_config.parallel_config
        scheduler_config = vllm_config.scheduler_config
        parallel_config.worker_cls = \
                        "tpu_inference.worker.tpu_worker.TPUWorker"

        multihost_backend = envs.TPU_MULTIHOST_BACKEND
        if not multihost_backend:  # Single host
            if parallel_config.pipeline_parallel_size == 1:
                logger.info("Force using UniProcExecutor for JAX on \
                        single host without pipeline parallelism.")
                parallel_config.distributed_executor_backend = "uni"
            else:
                logger.info("Force using MultiprocExecutor for JAX on \
                        single host with pipeline parallelism.")
                parallel_config.distributed_executor_backend = "mp"
        elif multihost_backend == "ray":
            from tpu_inference.executors.ray_distributed_executor import \
                RayDistributedExecutor
            parallel_config.distributed_executor_backend = RayDistributedExecutor
            logger.info(
                "Force using RayDistributedExecutor for JAX on multihost.")
        else:
            logger.warning(
                f"Unknown TPU multihost backend: {multihost_backend}. "
                "Using uniproc_executor.")
            parallel_config.distributed_executor_backend = "uni"

        if scheduler_config.is_multimodal_model and not \
            scheduler_config.disable_chunked_mm_input:
            logger.warning("TPU does not support running Multimodal models"\
            " without setting `--disable_chunked_mm_input`. " \
            "Forcing --disable_chunked_mm_input.")
            scheduler_config.disable_chunked_mm_input = True

        kv_transfer_config = vllm_config.kv_transfer_config
        if kv_transfer_config is not None:
            assert kv_transfer_config.kv_connector == "TPUConnector"
        # Late initialization to avoid circular import
        from tpu_inference.models.jax.utils.quantization.quantization_utils import \
            update_vllm_config_for_qwix_quantization

        update_vllm_config_for_qwix_quantization(vllm_config)

        from tpu_inference.core.sched.dp_scheduler import \
            update_vllm_config_for_dp_scheduler
        update_vllm_config_for_dp_scheduler(vllm_config)

    @classmethod
    def is_pin_memory_available(cls):
        logger.warning("Pin memory is not supported on TPU.")
        return False

    @classmethod
    def get_device_communicator_cls(cls) -> str:
        return "vllm.distributed.device_communicators.tpu_communicator.TpuCommunicator"  # noqa

    @classmethod
    def use_all_gather(cls) -> bool:
        return True

    @classmethod
    def supports_v1(cls, model_config: ModelConfig) -> bool:
        # V1 support on TPU is experimental
        return True

    @classmethod
    def validate_request(
        cls,
        prompt: PromptType,
        params: Union["SamplingParams", PoolingParams],
        processed_inputs: ProcessorInputs,
    ) -> None:
        """Raises if this request is unsupported on this platform"""
        from vllm.sampling_params import SamplingParams, SamplingType

        if isinstance(params, SamplingParams):
            if params.sampling_type == SamplingType.RANDOM_SEED:
                raise ValueError("JAX does not support per-request seed.")

    @classmethod
    def is_kv_cache_dtype_supported(cls, kv_cache_dtype: str,
                                    model_config: ModelConfig) -> bool:
        return True

    @classmethod
    def use_sync_weight_loader(cls) -> bool:
        """
        Returns if the current platform needs to sync weight loader.
        """
        return True

    @classmethod
    def support_hybrid_kv_cache(cls) -> bool:
        return True



import torchax
from jax.sharding import Mesh, PartitionSpec
from vllm.config import VllmConfig
from vllm.logger import init_logger
from vllm.model_executor.layers.fused_moe.layer import FusedMoE, FusedMoEConfig
# yapf: disable
from vllm.model_executor.layers.linear import (ColumnParallelLinear,
                                               LinearBase,
                                               MergedColumnParallelLinear,
                                               QKVParallelLinear,
                                               ReplicatedLinear,
                                               RowParallelLinear)

from tpu_inference.layers.vllm.linear_common import \
    get_model_matmul_fusion_assignment
from tpu_inference.utils import TPU_SECOND_LAST_MINOR

# yapf: enable

P = PartitionSpec

logger = init_logger(__name__)


class JaxCommonLinearConfig:

    def __init__(self, vllm_config: VllmConfig, mesh: Mesh, layer: LinearBase):
        assert isinstance(layer, LinearBase)

        self.mesh = mesh
        self.output_sizes = [layer.output_size]
        self.weight_sharding = P(None, None)
        self.fuse_matmuls = True
        self.enable_sp = vllm_config.compilation_config.pass_config.enable_sp
        self.input_sharding = None
        self.output_sharding = None

        if isinstance(layer, RowParallelLinear):
            self.weight_sharding = P(None, "model")
            if self.enable_sp:
                self.output_sharding = P("model", None)
        elif isinstance(layer, ColumnParallelLinear):
            self.weight_sharding = P("model", None)
            if self.enable_sp:
                self.input_sharding = P("model", None)

            if isinstance(layer, MergedColumnParallelLinear) or isinstance(
                    layer, QKVParallelLinear):
                self.output_sizes = layer.output_sizes

            self.fuse_matmuls = get_model_matmul_fusion_assignment(
                vllm_config.model_config.model,
                vllm_config.scheduler_config.max_num_batched_tokens,
                vllm_config.parallel_config.tensor_parallel_size,
                layer._get_name())
        elif isinstance(layer, ReplicatedLinear):
            self.weight_sharding = P(None, None)
        else:
            logger.warning(
                "Unsupported linear layer type of %s. Can potentially yield "
                " bad performance.", type(layer))

        self.bias_sharding = P(self.weight_sharding[0])
        if isinstance(self.weight_sharding[0], tuple):
            self.n_shards = 1
            for axis in self.weight_sharding[0]:
                self.n_shards *= self.mesh.shape.get(axis, 1)
        else:
            self.n_shards = self.mesh.shape.get(self.weight_sharding[0], 1)

    def get_input_sharding(self, x: torchax.tensor.Tensor):
        if self.enable_sp:
            token_num = x.shape[0]
            # NOTE(chengjiyao): make sure the sharded token_num is larger than TPU_SECOND_LAST_MINOR
            if token_num // self.mesh.shape["model"] >= TPU_SECOND_LAST_MINOR:
                return self.input_sharding
            else:
                return None
        return self.input_sharding

    def get_output_sharding(self, x: torchax.tensor.Tensor):
        if self.enable_sp:
            token_num = x.shape[0]
            # NOTE(chengjiyao): make sure the sharded token_num is larger than TPU_SECOND_LAST_MINOR
            if token_num // self.mesh.shape["model"] >= TPU_SECOND_LAST_MINOR:
                return self.output_sharding
            else:
                return None
        return self.output_sharding


class JaxCommonConfig:
    vllm_config: VllmConfig
    mesh: Mesh

    @classmethod
    def set_configs(cls, vllm_config: VllmConfig, mesh: Mesh):
        cls.vllm_config = vllm_config
        cls.mesh = mesh

    def get_linear_config(self, layer: LinearBase) -> JaxCommonLinearConfig:
        assert isinstance(layer, LinearBase)
        return JaxCommonLinearConfig(self.vllm_config, self.mesh, layer)

    def get_moe_config(self, layer: FusedMoE) -> FusedMoEConfig:
        assert isinstance(layer, FusedMoE)
        moe_config = layer.moe_config
        use_ep = self.vllm_config.parallel_config.enable_expert_parallel
        moe_config.moe_parallel_config.use_ep = use_ep
        return moe_config



import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized
from jax._src import dtypes
from jax._src import test_util as jtu

from tpu_inference.kernels.ragged_paged_attention.v3.kernel_hd64 import (
    ragged_paged_attention_hd64, ref_ragged_paged_attention_hd64)
from tpu_inference.kernels.ragged_paged_attention.v3.util import (
    align_to, cdiv, get_dtype_packing)

jax.config.parse_flags_with_absl()


@jtu.with_config(jax_numpy_dtype_promotion="standard")
class RaggedPagedAttentionHeadDim64KernelTest(jtu.JaxTestCase):

    def _test_ragged_paged_attention_hd64(
        self,
        seq_lens,  # List[(q_len, kv_len)]
        num_heads,  # [num_q_heads, num_kv_heads]
        head_dim,
        page_size,
        q_dtype,
        kv_dtype,
        num_pages,
        *,
        num_kv_pages_per_block=8,
        num_queries_per_block=64,
        vmem_limit_bytes=100 * 1024 * 1024,
        max_num_batched_tokens=512,
        max_num_seq=8,
        sliding_window: int | None = None,
        soft_cap: float | None = None,
        q_scale: float | None = None,
        k_scale: float | None = None,
        v_scale: float | None = None,
        use_attention_sink: bool = False,
    ):
        assert head_dim == 64
        rng = np.random.default_rng(1234)

        def gen_random(shape, dtype):
            return jnp.array(rng.random(size=shape,
                                        dtype=np.float32)).astype(dtype)

        if not jtu.is_device_tpu_at_least(version=4):
            self.skipTest("Expect TPUv4+")
        cu_q_lens = [0]
        kv_lens = []
        seq_lens = sorted(seq_lens, key=lambda x: x[0])
        num_decoding_seqs = sum(q_len for q_len, _ in seq_lens if q_len == 1)
        for q_len, kv_len in seq_lens:
            assert q_len <= kv_len
            cu_q_lens.append(cu_q_lens[-1] + q_len)
            kv_lens.append(kv_len)

        max_num_batched_tokens = max(align_to(cu_q_lens[-1], 128),
                                     max_num_batched_tokens)
        max_num_seq = max(align_to(len(seq_lens), 8), max_num_seq)
        max_kv_len = max(kv_lens)
        pages_per_seq = cdiv(max_kv_len, page_size)
        num_q_heads, num_kv_heads = num_heads

        q = gen_random((max_num_batched_tokens, num_q_heads, head_dim),
                       q_dtype)
        k = gen_random((max_num_batched_tokens, num_kv_heads, head_dim),
                       kv_dtype)
        v = gen_random((max_num_batched_tokens, num_kv_heads, head_dim),
                       kv_dtype)
        attention_sink = gen_random(
            (num_q_heads), jnp.float32) * 10**2 if use_attention_sink else None

        page_cnt = 0
        page_indices_list = []
        kv_pages_list = []
        kv_packing = get_dtype_packing(kv_dtype)
        padded_head_dim = align_to(head_dim, 128)
        padded_num_kv_heads = align_to(num_kv_heads, kv_packing)
        for kv_len in kv_lens:
            kv = gen_random(
                (
                    kv_len,
                    padded_num_kv_heads // kv_packing,
                    kv_packing,
                    padded_head_dim,
                ),
                kv_dtype,
            )
            kv = jnp.pad(
                kv,
                (
                    (
                        0,
                        cdiv(kv_len, page_size) * page_size - kv_len,
                    ),
                    (0, 0),
                    (0, 0),
                    (0, 0),
                ),
                constant_values=jnp.nan,
            ).reshape(
                -1,
                page_size,
                padded_num_kv_heads // kv_packing,
                kv_packing,
                padded_head_dim,
            )
            indices = page_cnt + jnp.arange(kv.shape[0], dtype=jnp.int32)
            indices = jnp.pad(
                indices,
                ((0, pages_per_seq - indices.shape[0]), ),
                constant_values=jnp.nan,
            )
            page_indices_list.append(indices)
            page_cnt += kv.shape[0]
            kv_pages_list.append(kv)

        kv_cache = jnp.concatenate(kv_pages_list, axis=0)
        kv_cache = jnp.pad(
            kv_cache,
            ((0, num_pages - kv_cache.shape[0]), (0, 0), (0, 0), (0, 0),
             (0, 0)),
            constant_values=jnp.nan,
        )
        page_indices = jnp.stack(page_indices_list, axis=0)
        page_indices = jnp.pad(
            page_indices,
            ((0, max_num_seq - page_indices.shape[0]), (0, 0)),
            constant_values=jnp.nan,
        )
        page_indices = page_indices.reshape(-1)

        cu_q_lens = jnp.array(cu_q_lens, dtype=jnp.int32)
        cu_q_lens = jnp.pad(cu_q_lens,
                            (0, max_num_seq + 1 - cu_q_lens.shape[0]))
        kv_lens = jnp.array(kv_lens, dtype=jnp.int32)
        kv_lens = jnp.pad(kv_lens, (0, max_num_seq - kv_lens.shape[0]))
        distribution = jnp.array(
            [num_decoding_seqs, num_decoding_seqs,
             len(seq_lens)],
            dtype=jnp.int32)

        args = (
            q,
            k,
            v,
            kv_cache,
            kv_lens,
            page_indices,
            cu_q_lens,
            distribution,
            attention_sink,
        )

        kwargs = {
            "sliding_window": sliding_window,
            "soft_cap": soft_cap,
            "q_scale": q_scale,
            "k_scale": k_scale,
            "v_scale": v_scale,
        }

        expected, expected_kv_cache = ref_ragged_paged_attention_hd64(
            *args,
            **kwargs,
        )

        output, updated_kv_cache = ragged_paged_attention_hd64(
            *args,
            **kwargs,
            num_kv_pages_per_block=num_kv_pages_per_block,
            num_queries_per_block=num_queries_per_block,
            vmem_limit_bytes=vmem_limit_bytes,
        )
        output = output[:cu_q_lens[distribution[-1]]]

        dtype_bits = (dtypes.bit_width(jnp.dtype(kv_dtype)) if hasattr(
            dtypes, "bit_width") else dtypes.itemsize_bits(
                jnp.dtype(kv_dtype)))
        tols = {
            32: 0.15,
            16: 0.2,
            8: 0.2,
            4: 0.2,
        }
        tol = tols[dtype_bits]
        self.assertAllClose(output, expected, atol=tol, rtol=tol)
        mask = ~jnp.isnan(expected_kv_cache)
        self.assertArraysEqual(updated_kv_cache[mask], expected_kv_cache[mask])
        self.assertEqual(output.shape[-1], head_dim)

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_basic(self, dtype):
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
        )

    # TODO: support integer (int8, int4) and fp4 kv cache
    @parameterized.product(
        q_dtype=[jnp.bfloat16],
        kv_dtype=[jnp.float8_e5m2, jnp.float8_e4m3fn],
        kv_scales=[(0.5, 0.5), (None, None)],
    )
    def test_ragged_paged_attention_quantized_kv_cache(self, q_dtype, kv_dtype,
                                                       kv_scales):
        if not jtu.is_device_tpu_at_least(version=5):
            self.skipTest("Expect TPUv5+")
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000
        k_scale, v_scale = kv_scales

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            q_dtype,
            kv_dtype,
            num_pages,
            k_scale=k_scale,
            v_scale=v_scale,
        )

    @parameterized.product(
        q_dtype=[jnp.bfloat16],
        kv_dtype=[jnp.float8_e5m2, jnp.float8_e4m3fn],
        q_scale=[0.5],
        kv_scales=[(0.5, 0.5), (None, None)],
    )
    def test_ragged_paged_attention_quantized_attention(
            self, q_dtype, kv_dtype, q_scale, kv_scales):
        if not jtu.is_device_tpu_at_least(version=5):
            self.skipTest("Expect TPUv5+")
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000
        k_scale, v_scale = kv_scales

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            q_dtype,
            kv_dtype,
            num_pages,
            q_scale=q_scale,
            k_scale=k_scale,
            v_scale=v_scale,
        )

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_decode_only(self, dtype):
        seq_lens = [
            (1, 18),
            (1, 129),
            (1, 597),
            (1, 122),
            (1, 64),
            (1, 322),
            (1, 463),
            (1, 181),
            (1, 1107),
            (1, 123),
            (1, 31),
            (1, 18),
            (1, 1229),
            (1, 229),
            (1, 87),
            (1, 1328),
        ]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
        )

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_prefill_only(self, dtype):
        seq_lens = [
            (5, 18),
            (15, 129),
            (120, 597),
            (100, 122),
            (21, 64),
            (32, 322),
            (251, 463),
            (40, 181),
            (64, 1107),
            (99, 123),
            (10, 31),
            (5, 18),
            (3, 1229),
            (120, 229),
            (9, 87),
            (2, 1328),
        ]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
        )

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_mixed(self, dtype):
        seq_lens = [
            (5, 18),
            (1, 129),
            (120, 597),
            (1, 122),
            (1, 64),
            (32, 322),
            (251, 463),
            (1, 181),
            (1, 1107),
            (99, 123),
            (1, 31),
            (5, 18),
            (3, 1229),
            (117, 229),
            (1, 87),
            (1, 1328),
        ]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
        )

    @parameterized.product(
        num_seqs=[1, 17],
        num_heads=[(32, 8), (12, 2), (5, 1), (3, 3)],
        head_dim=[64],
        dtype=[jnp.float32, jnp.bfloat16],
        # num_kv_pages_per_block=[8, 16],
        # num_queries_per_block=[16, 32],
    )
    def test_ragged_paged_attention_complex(
        self,
        num_seqs,
        num_heads,
        head_dim,
        dtype,
        # num_kv_pages_per_block,
        # num_queries_per_block,
    ):
        rng = np.random.default_rng(1234)
        q_lens = rng.integers(1, 100, num_seqs)
        kv_lens = q_lens + rng.integers(0, 50, num_seqs)
        seq_lens = list(zip(q_lens.tolist(), kv_lens.tolist()))
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
            # num_kv_pages_per_block=num_kv_pages_per_block,
            # num_queries_per_block=num_queries_per_block,
        )

    @parameterized.product(sliding_window=[None, 5, 128], )
    def test_ragged_paged_attention_sliding_window(
        self,
        sliding_window: int | None,
    ):
        num_seqs = 5
        num_heads = (4, 4)
        dtype = jnp.float32
        rng = np.random.default_rng(1234)
        q_lens = rng.integers(1, 100, num_seqs)
        kv_lens = q_lens + rng.integers(0, 50, num_seqs)
        seq_lens = list(zip(q_lens.tolist(), kv_lens.tolist()))
        head_dim = 64
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
            sliding_window=sliding_window,
        )

    @parameterized.product(
        sliding_window=[5, 128],
        num_heads=[(4, 4), (8, 4), (64, 8)],
    )
    def test_ragged_paged_attention_sliding_window_with_attention_sink_hd64(
        self,
        sliding_window: int | None,
        num_heads: tuple[int, int],
    ):
        num_seqs = 5
        dtype = jnp.float32
        rng = np.random.default_rng(1234)
        q_lens = rng.integers(1, 100, num_seqs)
        kv_lens = q_lens + rng.integers(0, 50, num_seqs)
        seq_lens = list(zip(q_lens.tolist(), kv_lens.tolist()))
        head_dim = 64
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
            sliding_window=sliding_window,
            use_attention_sink=True,
        )

    @parameterized.product(soft_cap=[None, 50.0], )
    def test_ragged_paged_attention_logit_soft_capping(
        self,
        soft_cap: float | None,
    ):
        num_heads = (16, 2)
        num_seqs = 2
        dtype = jnp.float32
        rng = np.random.default_rng(1234)
        q_lens = rng.integers(1, 100, num_seqs)
        kv_lens = q_lens + rng.integers(0, 50, num_seqs)
        seq_lens = list(zip(q_lens.tolist(), kv_lens.tolist()))
        head_dim = 64
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention_hd64(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            dtype,
            num_pages,
            soft_cap=soft_cap,
        )

    def test_ragged_paged_attention_sliding_window_should_be_positive(self):
        dtype = jnp.float32
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000

        with self.assertRaisesRegex(ValueError, "must be positive"):
            self._test_ragged_paged_attention_hd64(
                seq_lens,
                num_heads,
                head_dim,
                page_size,
                dtype,
                dtype,
                num_pages,
                sliding_window=0,
            )

        with self.assertRaisesRegex(ValueError, "must be positive"):
            self._test_ragged_paged_attention_hd64(
                seq_lens,
                num_heads,
                head_dim,
                page_size,
                dtype,
                dtype,
                num_pages,
                sliding_window=-1,
            )

    def test_ragged_paged_attention_soft_cap_cannot_be_zero(self):
        dtype = jnp.float32
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 64
        page_size = 16
        num_pages = 1000

        with self.assertRaisesRegex(ValueError, "must not be 0.0"):
            self._test_ragged_paged_attention_hd64(
                seq_lens,
                num_heads,
                head_dim,
                page_size,
                dtype,
                dtype,
                num_pages,
                soft_cap=0.0,
            )


if __name__ == "__main__":
    absltest.main(testLoader=jtu.JaxTestLoader())



from dataclasses import InitVar, dataclass

import jax
import jax.numpy as jnp
from flax import nnx
from flax.typing import Sharding
from jaxtyping import Float

from tpu_inference.layers.jax.base import create_param
from tpu_inference.layers.jax.layers import FlaxUtils
from tpu_inference.layers.jax.moe.moe import Router

modeling_flax_utils = FlaxUtils()


@dataclass(kw_only=True)
class GptOssRouter(Router):
    """Router module for Mixture-of-Experts (MoE) layers.

    This module determines which experts each token should be routed.

    """
    e_sharding: Sharding = ()

    def __post_init__(self, rngs: nnx.Rngs):
        """
        Initializes the parent's kernel and adds the new bias parameter.
        """
        super().__post_init__(rngs)

        self.bias_E = create_param(rngs,
                                   shape=(self.num_experts, ),
                                   dtype=self.dtype,
                                   sharding=self.e_sharding,
                                   random_init=self.random_init)

    def __call__(self, x_TD: Float):
        """
        Overrides the parent's forward pass to include the bias.
        """
        x_TD = jnp.asarray(x_TD, self.dtype)
        x_TD = nnx.with_sharding_constraint(x_TD, self.activation_ffw_td)

        router_logits_TE = jnp.einsum('TD,DE -> TE', x_TD,
                                      self.kernel_DE.value)

        router_logits_TE += self.bias_E.value

        weights_TX, selected_experts_TX = jax.lax.top_k(
            router_logits_TE, self.num_experts_per_tok)

        normalized_weights_TX = jax.nn.softmax(weights_TX.astype(self.dtype),
                                               axis=-1)

        return normalized_weights_TX, selected_experts_TX


def _swiglu(x: Float, alpha: Float, limit: Float) -> Float:
    """Implements the specific SwiGLU from the golden implementation."""
    x_glu, x_linear = x[..., ::2], x[..., 1::2]

    x_glu = jnp.clip(x_glu, a_max=limit)
    x_linear = jnp.clip(x_linear, a_min=-limit, a_max=limit)

    gated_activation = x_glu * jax.nn.sigmoid(alpha * x_glu)

    return gated_activation * (x_linear + 1)


@dataclass(kw_only=True)
class CombineExperts(nnx.Module):
    """Module for combining expert outputs with weighted sum."""
    dtype: jnp.dtype

    def __call__(self, down_proj_TED: Float, weights_TX: Float,
                 indices_TX: jax.Array) -> Float:
        """Combines expert outputs using weighted sum.

        Args:
            down_proj_TED: Expert outputs, shape (tokens, experts, hidden_dim)
            weights_TX: Router weights, shape (tokens, experts_per_token)
            indices_TX: Selected expert indices, shape (tokens, experts_per_token)

        Returns:
            Combined output, shape (tokens, hidden_dim)
        """
        with jax.named_scope("combine_experts"):
            indices_for_gather = indices_TX[..., None]
            gathered_down_proj_TED = jnp.take_along_axis(down_proj_TED,
                                                         indices_for_gather,
                                                         axis=1)
            output_TD = jnp.einsum('TXD,TX -> TD', gathered_down_proj_TED,
                                   weights_TX)

        return output_TD.astype(self.dtype)


@dataclass(kw_only=True)
class GptOssMoE(nnx.Module):
    """
    JAX implementation of the GPT-OSS Mixture-of-Experts MLP block.
    """
    dtype: jnp.dtype
    hidden_size: int
    intermediate_size_moe: int
    num_local_experts: int
    router: GptOssRouter
    rngs: InitVar[nnx.Rngs]

    swiglu_limit: float = 7.0
    swiglu_alpha: float = 1.702

    # Sharding specifications
    activation_ffw_td: Sharding
    edf_sharding: Sharding
    efd_sharding: Sharding
    ed_sharding: Sharding

    random_init: bool = False

    def __call__(self, x_TD: Float) -> Float:
        """Performs the forward pass for the GPT-OSS MoE layer."""
        x_TD = jnp.asarray(x_TD, self.dtype)
        x_TD = nnx.with_sharding_constraint(x_TD, self.activation_ffw_td)

        weights_TX, indices_TX = self.router(x_TD)

        # First MLP layer (up-projection)
        with jax.named_scope("MLP #1"):
            up_proj_TEF2 = jnp.einsum('TD,EDF -> TEF', x_TD,
                                      self.mlp1_weight_EDF2.value)
            up_proj_TEF2 += self.mlp1_bias_EF2.value

            fuse_TEF = _swiglu(up_proj_TEF2,
                               alpha=self.swiglu_alpha,
                               limit=self.swiglu_limit)

        # Second MLP layer (down-projection)
        with jax.named_scope("MLP #2"):
            down_proj_TED = jnp.einsum('TEF,EFD -> TED', fuse_TEF,
                                       self.mlp2_weight_EFD.value)
            down_proj_TED += self.mlp2_bias_ED.value

        # Weighted sum of expert outputs
        output_TD = self.combine_experts(down_proj_TED, weights_TX, indices_TX)

        return output_TD

    def __post_init__(self, rngs: nnx.Rngs):
        """Initializes all weights and biases for the MoE block."""
        D, F, E = self.hidden_size, self.intermediate_size_moe, self.num_local_experts

        self.combine_experts = CombineExperts(dtype=self.dtype)

        # MLP #1 Weights (Combined Gate and Up-projection) and Bias
        self.mlp1_weight_EDF2 = create_param(
            rngs,
            shape=(E, D, F * 2),
            dtype=self.dtype,
            sharding=self.edf_sharding,
            random_init=self.random_init,
        )
        self.mlp1_bias_EF2 = create_param(
            rngs,
            shape=(E, F * 2),
            dtype=self.dtype,
            sharding=self.ed_sharding,
            random_init=self.random_init,
        )

        # MLP #2 Weights (Down-projection) and Bias
        self.mlp2_weight_EFD = create_param(
            rngs,
            shape=(E, F, D),
            dtype=self.dtype,
            sharding=self.efd_sharding,
            random_init=self.random_init,
        )
        self.mlp2_bias_ED = create_param(
            rngs,
            shape=(E, D),
            dtype=self.dtype,
            sharding=self.ed_sharding,
            random_init=self.random_init,
        )



import enum
from dataclasses import InitVar, dataclass
from functools import partial
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
from flax import nnx
from flax.typing import Sharding
from jax.sharding import PartitionSpec
from jaxtyping import Float
from qwix._src.core.ragged_dot import ragged_dot as qwix_ragged_dot
from qwix._src.providers import ptq

from tpu_inference.layers.jax.base import create_param
from tpu_inference.layers.jax.layers import FlaxUtils
from tpu_inference.layers.jax.moe.moe import MoE
from tpu_inference.models.jax.utils.quantization.quantization_utils import (
    manually_quantize_qwix_activation, manually_quantize_qwix_weight)

modeling_flax_utils = FlaxUtils()


@dataclass
class DeepSeekV3Router(nnx.Module):
    """Router module for Mixture-of-Experts (MoE) layers.

    This module determines which experts each token should be routed to based on the input.

    """

    hidden_size: int
    num_experts: int
    num_experts_per_tok: int
    n_groups: int
    topk_groups: int
    norm_topk_prob: bool
    routed_scaling_factor: float
    dtype: jnp.dtype
    rngs: InitVar[nnx.Rngs]

    # Sharding Attributes
    activation_ffw_td: Sharding = ()
    ed_sharding: Sharding = ()
    e_sharding: Sharding = ()

    random_init: bool = False

    router_bias_dtype: jnp.dtype = jnp.float32

    def get_topk_indices(self, scores_TE: Float) -> Float:
        """Get the topk indices of the scores.

        Args:
            scores_TE: The scores to get the topk indices of. Shape (sequence, num_experts).

        Returns:
            The topk indices of the scores. Shape (sequence, num_experts_per_tok).
        """

        scores_TE = scores_TE + self.bias_E
        if self.n_groups > 1:
            experts_per_group = self.num_experts // self.n_groups
            group_scores_TGM = jnp.reshape(
                scores_TE, (-1, self.n_groups, experts_per_group))
            group_scores_TG2 = jax.lax.top_k(group_scores_TGM, k=2)[0]
            group_scores_TG = jnp.sum(group_scores_TG2, axis=-1)
            indices = jax.lax.top_k(group_scores_TG, k=self.topk_groups)[1]

            mask_TG = jnp.any(jnp.arange(
                self.n_groups)[:, None] == indices[..., None, :],
                              axis=-1)
            mask_TE = jnp.repeat(mask_TG,
                                 scores_TE.shape[-1] // mask_TG.shape[-1], -1)
            scores_TE = jnp.where(mask_TE, scores_TE, 0.0)

        indices_TX = jax.lax.top_k(scores_TE, k=self.num_experts_per_tok)[1]

        return indices_TX

    def __call__(self, x_TD: Float) -> Tuple[Float, Float]:
        """Routes tokens to top k experts.

        Args:
            x_TD: Input array of shape (sequence, d_model).

        Returns:
            A tuple containing:
                - weights: Normalized weights for selected experts, shape (sequence, num_experts_per_tok).
                - indices: Indices of selected experts, shape (sequence, num_experts_per_tok).
        """
        x_TD = jnp.asarray(x_TD, self.dtype)
        x_TD = nnx.with_sharding_constraint(x_TD, self.activation_ffw_td)

        scores_TE = jnp.einsum("TD,DE -> TE", x_TD, self.kernel_DE.value)
        scores_TE = nnx.sigmoid(scores_TE)

        original_scores_TE = scores_TE
        topk_indices_TX = self.get_topk_indices(scores_TE)
        weights_TX = jnp.take_along_axis(original_scores_TE,
                                         topk_indices_TX,
                                         axis=-1)

        if self.norm_topk_prob:
            weights_TX /= jnp.sum(weights_TX, axis=-1)[..., None] + 1e-20

        weights_TX *= self.routed_scaling_factor

        return weights_TX, topk_indices_TX

    def __post_init__(self, rngs: nnx.Rngs):
        """Generates the router kernel (weights and bias) for routing."""
        D = self.hidden_size
        E = self.num_experts
        self.kernel_DE = create_param(rngs,
                                      shape=(D, E),
                                      dtype=self.dtype,
                                      sharding=self.ed_sharding,
                                      random_init=self.random_init)
        self.bias_E = create_param(rngs,
                                   shape=(E, ),
                                   dtype=self.router_bias_dtype,
                                   sharding=self.e_sharding,
                                   random_init=self.random_init)


@dataclass(kw_only=True)
class SparseMoE(MoE):
    """Mixture-of-Experts (MoE) Routed MLP Layer.

    This module implements a Sparse MoE layer with a router and multiple expert MLPs.

    Attributes:
        num_experts_per_tok: The number of experts each token is routed to.
        tile_size: A tuple (batch, activation_dim, weight_dim) for GMM tiling.
        use_megablox: If True, uses the MegaBlox GMM kernel.
        mesh: The device mesh.
        # TODO: need to redesign this I/O for parallelism
        num_expert_parallelism: The size of the 'expert' mesh dimension.
        # TODO: determine if we get it from external or extrat it in MoE class
        is_batch_sharded_by_expert: True if batch is sharded over 'expert' dim.
    """
    num_experts_per_tok: int
    #TODO: tile size is (tile_batch_seq, tile_activation_dim, tile_weight_dim,) from MaxText
    tile_size: tuple[int, int, int] = (128, 64, 128)
    use_megablox: bool = False
    mesh: jax.sharding.Mesh
    # This should be set if and only if you have quantized your model (via Qwix)
    quantized_dtype: Optional[jnp.dtype] = None

    def __post_init__(self, rngs: nnx.Rngs):
        super().__post_init__(rngs)

        # Derive the expert sharding
        self.expert_axis_name = self.edf_sharding[0]
        if self.expert_axis_name is None:
            self.num_expert_parallelism = 1
        else:
            self.num_expert_parallelism = self.mesh.shape[
                self.expert_axis_name]

        # Derive if data is sharded by expert
        self.data_axis_name = self.activation_ffw_td[0]
        self.is_batch_sharded_by_expert = (
            self.expert_axis_name is not None) and (self.expert_axis_name
                                                    == self.data_axis_name)

    def _sort_activations(self, inputs: jax.Array,
                          sort_indices: jax.Array) -> jax.Array:
        """Sorts activations(inputs) by `sort_indices` for the forward pass."""
        return inputs[sort_indices, ...]

    @staticmethod
    def get_all_to_all_params(
        all_shards_group_sizes,
        shard_id,
        num_expert_parallelism,
        is_batch_sharded=True,
    ):
        """Generates params for ragged_all_to_all communication."""

        class TransformStrategy(enum.Enum):
            INPUT_OFFSET = enum.auto()
            SEND_SIZE = enum.auto()
            OUTPUT_OFFSET = enum.auto()
            RECV_SIZE = enum.auto()

        def transform_array(input_array, shard_id, strategy, is_batch_sharded):
            if is_batch_sharded:
                if strategy == TransformStrategy.INPUT_OFFSET:
                    local_array = input_array[shard_id]
                    return jnp.concatenate(
                        (jnp.array([0]), jnp.cumsum(local_array)[:-1]))
                elif strategy == TransformStrategy.SEND_SIZE:
                    return input_array[shard_id]
                elif strategy == TransformStrategy.OUTPUT_OFFSET:
                    zero_row = jnp.zeros((1, ) + input_array.shape[1:],
                                         dtype=input_array.dtype)
                    array_with_zeros = jnp.concatenate((zero_row, input_array),
                                                       axis=0)
                    cumulated_array = jnp.cumsum(array_with_zeros,
                                                 axis=0,
                                                 dtype=input_array.dtype)
                    return cumulated_array[shard_id]
                elif strategy == TransformStrategy.RECV_SIZE:
                    return input_array[:, shard_id]
                else:
                    raise ValueError(
                        f"Unknown transform array strategy: {strategy}")
            else:
                if strategy == TransformStrategy.INPUT_OFFSET:
                    return jnp.zeros(num_expert_parallelism,
                                     dtype=input_array.dtype)
                elif strategy == TransformStrategy.SEND_SIZE:
                    return jnp.repeat(input_array[shard_id],
                                      num_expert_parallelism)
                elif strategy == TransformStrategy.OUTPUT_OFFSET:
                    output_offset = jnp.concatenate(
                        (jnp.array([0]),
                         jnp.cumsum(input_array[:-1])))[shard_id]
                    return jnp.repeat(output_offset, num_expert_parallelism)
                elif strategy == TransformStrategy.RECV_SIZE:
                    return input_array
                else:
                    raise ValueError(
                        f"Unknown transform array strategy: {strategy}")

        input_offsets = transform_array(all_shards_group_sizes, shard_id,
                                        TransformStrategy.INPUT_OFFSET,
                                        is_batch_sharded)
        send_sizes = transform_array(all_shards_group_sizes, shard_id,
                                     TransformStrategy.SEND_SIZE,
                                     is_batch_sharded)
        output_offsets = transform_array(all_shards_group_sizes, shard_id,
                                         TransformStrategy.OUTPUT_OFFSET,
                                         is_batch_sharded)
        recv_sizes = transform_array(all_shards_group_sizes, shard_id,
                                     TransformStrategy.RECV_SIZE,
                                     is_batch_sharded)
        return input_offsets, send_sizes, output_offsets, recv_sizes

    def _local_permute(
        self,
        inputs,
        global_group_sizes,
        local_expert_size,
        shard_index,
        is_offset=False,
        global_sorted_experts=None,
    ):
        """Permutes tokens locally within an expert shard."""
        # global_group_sizes: (tokens parallelism, num_total_experts)
        # all_shard_local_sizes: (tokens parallelism, num local experts in the shard)
        all_shard_local_sizes = jax.lax.dynamic_slice_in_dim(
            global_group_sizes,
            shard_index * local_expert_size,
            local_expert_size,
            axis=1,
        )
        local_sizes = all_shard_local_sizes.reshape(-1)

        # local_group_size: (tokens parallelism, )
        local_group_size = jnp.sum(all_shard_local_sizes, axis=0)

        # When token replicated in devices
        if is_offset:
            global_sorted_shard_assignments = jnp.floor_divide(
                global_sorted_experts, local_expert_size)
            expert_indices = jnp.where(
                global_sorted_shard_assignments == shard_index,
                jnp.mod(global_sorted_experts, local_expert_size),
                local_expert_size,
            )

        # When token sharded in devices
        else:
            base_indices = jnp.mod(jnp.arange(local_sizes.shape[0]),
                                   local_expert_size)
            expert_indices = jnp.repeat(base_indices,
                                        local_sizes,
                                        total_repeat_length=inputs.shape[0])

        sorted_indices = jnp.argsort(expert_indices)
        # sort the inputs based on the local expert_indices
        sorted_inputs = self._sort_activations(inputs, sorted_indices)
        # sortted local expert id from 0 to local expert size
        sorted_experts_ids = expert_indices[sorted_indices]
        return (
            sorted_inputs,
            sorted_indices,
            local_group_size,
            sorted_experts_ids,
        )

    def _permute(self, inputs_TD: Float, selected_experts_TX: jax.Array):
        """Global permute: Sorts tokens by assigned expert."""
        # suffix t = T * X = total_assignments for the local tokens(T) on this device.
        total_tokens = inputs_TD.shape[0]
        flat_expert_indices = selected_experts_TX.flatten()
        sort_indices_t = jnp.argsort(flat_expert_indices)

        replicated_inputs_tD = jnp.repeat(inputs_TD,
                                          self.num_experts_per_tok,
                                          axis=0)
        sorted_inputs_tD = self._sort_activations(replicated_inputs_tD,
                                                  sort_indices_t)

        # number of tokens assigned to each expert
        group_sizes_E = jnp.bincount(flat_expert_indices,
                                     length=self.num_local_experts)

        expert_ids = jnp.arange(self.num_local_experts)
        total_assignments = total_tokens * self.num_experts_per_tok
        sorted_expert_assignments_t = jnp.repeat(
            expert_ids,
            repeats=group_sizes_E,
            total_repeat_length=total_assignments)

        return (
            sorted_inputs_tD,
            sort_indices_t,
            group_sizes_E,
            sorted_expert_assignments_t,
        )

    def _unpermute(self, processed_tokens: jax.Array, sort_indices: jax.Array,
                   router_weights_TX: jax.Array):
        """Unsorts tokens to their original order and combines expert outputs with router's weight."""
        with jax.named_scope("unpermute"):
            unsorted_tokens_tD = self._sort_activations(
                processed_tokens, jnp.argsort(sort_indices))
            reshaped_tokens_TXD = unsorted_tokens_tD.reshape(
                -1, self.num_experts_per_tok, self.hidden_size)
        with jax.named_scope("combine_weights"):
            output_TD = jnp.einsum(
                "TXD,TX -> TD",
                reshaped_tokens_TXD.astype(jnp.float32),
                router_weights_TX.astype(jnp.float32),
                precision='float32',
            )

        return output_TD.astype(self.dtype)

    def _gmm(self, inputs, kernel, group_sizes):
        """Performs Grouped Matrix Multiply."""
        num_rows = inputs.shape[0]
        pad_amount = (self.tile_size[0] -
                      num_rows % self.tile_size[0]) % self.tile_size[0]
        if pad_amount > 0:
            inputs = jnp.pad(inputs, ((0, pad_amount), (0, 0)))

        if self.use_megablox:
            #TODO: megablox is used in MaxText, keep a placeholder here for future implement
            raise NotImplementedError(
                "MegaBlox kernel call is not implemented.")
        else:
            inputs = manually_quantize_qwix_activation(
                inputs, "ragged_dot", jnp.float8_e4m3fn, [0], {},
                "absmax") if self.quantized_dtype else inputs
            ragged_dot_func = qwix_ragged_dot if self.quantized_dtype else jax.lax.ragged_dot
            output = ragged_dot_func(
                lhs=inputs,
                rhs=kernel,
                group_sizes=group_sizes,
                preferred_element_type=self.dtype,
            )

        if pad_amount > 0:
            output = output[:num_rows, :]
        return output

    @staticmethod
    def _distributed_sparse_moe_fwd(
        self,
        x_TD: jax.Array,
        router_weights_TX: jax.Array,
        selected_experts_TX: jax.Array,
        kernel_gating: jax.Array,
        kernel_up_proj: jax.Array,
        kernel_down_proj: jax.Array,
    ):
        """
        The sparse MoE forward pass with fully distributed logic.
        This assumes it is running within a distributed TPU.
        """

        # 1. Global Permute, perpute all tokens across shards
        (
            sorted_inputs,
            global_sort_indices,
            global_group_sizes,
            global_sorted_experts,
        ) = self._permute(x_TD, selected_experts_TX)

        # TODO: update to 'expert' after we enable expert parallelism, currently experts are sharded along model axis
        # or we sould derive it from the model init
        expert_shard_id = jax.lax.axis_index(self.expert_axis_name)
        local_expert_size = self.num_local_experts // self.num_expert_parallelism

        if self.num_expert_parallelism > 1:
            if self.is_batch_sharded_by_expert:
                # When token sharded in devices
                # In this path, we assume the data(tokens) are fully sharded on expert, namely data_axis_name == expert_axis_name

                # 2a. Send Tokens To Experts (All-to-All)
                # Gather group sizes from all data shards
                # all_shards_group_sizes: (data parallelism = expert parallelism, number of total experts )
                all_shards_group_sizes = jax.lax.all_gather(
                    global_group_sizes, axis_name=self.data_axis_name)

                # all_shards_group_sizes_per_expert_shard[i][j] = # tokens on shard[i] to be sent to expert shard[j]
                all_shards_group_sizes_per_expert_shard = jnp.sum(
                    all_shards_group_sizes.reshape(
                        self.num_expert_parallelism,  # data parallelism
                        self.num_expert_parallelism,  # expert parallelism
                        local_expert_size  # Experts per shard
                    ),
                    axis=2)
                input_offsets, send_sizes, output_offsets, recv_sizes = self.get_all_to_all_params(
                    all_shards_group_sizes_per_expert_shard, expert_shard_id,
                    self.num_expert_parallelism)
                # Estimate buffer size
                local_total_assignments = x_TD.shape[
                    0] * self.num_experts_per_tok
                global_total_assignments = local_total_assignments * self.num_expert_parallelism
                output_shape_est = jnp.zeros(
                    (global_total_assignments, self.hidden_size),
                    dtype=sorted_inputs.dtype)

                inputs_after_all2all = jax.lax.ragged_all_to_all(
                    sorted_inputs,
                    output_shape_est,
                    input_offsets,
                    send_sizes,
                    output_offsets,
                    recv_sizes,
                    axis_name=self.expert_axis_name)

                # 3a. Local Permute
                # Get full group sizes from all shards
                full_global_group_sizes = jax.lax.all_gather(
                    global_group_sizes, axis_name=self.expert_axis_name)
                (
                    compute_inputs,
                    local_sorted_indices,
                    compute_group_sizes,
                    compute_expert_ids,
                ) = self._local_permute(
                    inputs_after_all2all,
                    full_global_group_sizes,
                    local_expert_size,
                    shard_index=expert_shard_id,
                    is_offset=False,
                )

            else:
                # When token replicated in devices

                # 2. No send all-to-all needed, as the tokens are sorted and replicated on all devices
                # 3b. Local "Permute"
                (
                    compute_inputs,
                    local_sorted_indices,
                    compute_group_sizes,
                    compute_expert_ids,
                ) = self._local_permute(
                    sorted_inputs,
                    global_group_sizes[None, :],
                    local_expert_size,
                    shard_index=expert_shard_id,
                    is_offset=True,
                    global_sorted_experts=global_sorted_experts,
                )

                # Calculate group sizes for return all-to-all
                reshaped_group_sizes = jnp.sum(global_group_sizes.reshape(
                    -1, local_expert_size),
                                               axis=1)
                mask = compute_expert_ids < local_expert_size
                compute_inputs = compute_inputs * mask[..., None]

        else:
            # --- NO EXPERT PARALLELISM ---
            compute_inputs = sorted_inputs
            compute_group_sizes = global_group_sizes
            compute_expert_ids = global_sorted_experts
            local_sorted_indices = jnp.arange(sorted_inputs.shape[0])

        # 4. Compute: Apply experts using Grouped Matrix Multiply
        with jax.named_scope("gating"):
            # compute_inputs: (local total assignments, D)
            gating_TEF = self._gmm(compute_inputs, kernel_gating,
                                   compute_group_sizes)
            activated_gating_TEF = modeling_flax_utils.ACT2FN[self.hidden_act](
                gating_TEF)

        with jax.named_scope("up_projection"):
            up_proj_TEF = self._gmm(compute_inputs, kernel_up_proj,
                                    compute_group_sizes)

        fuse_TEF = activated_gating_TEF * up_proj_TEF

        with jax.named_scope("down_projection"):
            # intermediate_output: (local total assignments, D)
            intermediate_output = self._gmm(fuse_TEF, kernel_down_proj,
                                            compute_group_sizes)

        # 5. Return Results (All-to-All)
        if self.num_expert_parallelism > 1:
            local_total_assignments = x_TD.shape[0] * self.num_experts_per_tok
            output_shape = jnp.zeros(
                (local_total_assignments, self.hidden_size),
                dtype=intermediate_output.dtype)

            if self.is_batch_sharded_by_expert:
                # When token sharded in devices
                # Unsort locally before sending back
                local_output = self._sort_activations(
                    intermediate_output, jnp.argsort(local_sorted_indices))

                input_offsets, send_sizes, output_offsets, recv_sizes = self.get_all_to_all_params(
                    jnp.transpose(all_shards_group_sizes),
                    expert_shard_id,
                    self.num_expert_parallelism,
                )
                final_intermediate_output = jax.lax.ragged_all_to_all(
                    local_output,
                    output_shape,
                    input_offsets,
                    send_sizes,
                    output_offsets,
                    recv_sizes,
                    axis_name=self.expert_axis_name)
            else:
                # When token replicated in devices
                input_offsets, send_sizes, output_offsets, recv_sizes = self.get_all_to_all_params(
                    reshaped_group_sizes,
                    expert_shard_id,
                    self.num_expert_parallelism,
                    is_batch_sharded=False,
                )
                final_intermediate_output = jax.lax.ragged_all_to_all(
                    intermediate_output,
                    output_shape,
                    input_offsets,
                    send_sizes,
                    output_offsets,
                    recv_sizes,
                    axis_name=self.expert_axis_name)
        else:
            final_intermediate_output = intermediate_output

        # 6. Global Unpermute (on the data shard)
        with jax.named_scope("unpermute"):
            output_TD = self._unpermute(final_intermediate_output,
                                        global_sort_indices, router_weights_TX)

        return output_TD

    def __call__(self, x_TD: Float):
        """Performs the forward pass of the Sparse MoE layer."""
        x_TD = jnp.asarray(x_TD, self.dtype)
        x_TD = nnx.with_sharding_constraint(x_TD, self.activation_ffw_td)
        router_weights_TX, selected_experts_TX = self.router(x_TD)

        in_specs = (
            PartitionSpec(),  # Replicated `self`
            PartitionSpec(*self.activation_ffw_td),  # Sharded x_TD
            PartitionSpec(),  # Replicated router_weights_TX
            PartitionSpec(),  # Replicated selected_experts_TX
            PartitionSpec(*self.edf_sharding),  # Sharded gating kernel
            PartitionSpec(*self.edf_sharding),  # Sharded up-projection kernel
            PartitionSpec(
                *self.efd_sharding),  # Sharded down-projection kernel
        )
        out_specs = PartitionSpec(*self.activation_ffw_td)

        mapped_moe_fwd = partial(jax.experimental.shard_map.shard_map,
                                 mesh=self.mesh,
                                 in_specs=in_specs,
                                 out_specs=out_specs,
                                 check_rep=False)(
                                     SparseMoE._distributed_sparse_moe_fwd)

        kernel_gating_EDF = self.kernel_gating_EDF.value
        kernel_up_proj_EDF = self.kernel_up_proj_EDF.value
        kernel_down_proj_EFD = self.kernel_down_proj_EFD.value

        if self.quantized_dtype:
            if not isinstance(kernel_gating_EDF, ptq.WithAux):
                kernel_gating_EDF = manually_quantize_qwix_weight(
                    kernel_gating_EDF, self.quantized_dtype, [0, 2], {},
                    "absmax")
            if not isinstance(kernel_up_proj_EDF, ptq.WithAux):
                kernel_up_proj_EDF = manually_quantize_qwix_weight(
                    kernel_up_proj_EDF, self.quantized_dtype, [0, 2], {},
                    "absmax")
            if not isinstance(kernel_down_proj_EFD, ptq.WithAux):
                kernel_down_proj_EFD = manually_quantize_qwix_weight(
                    kernel_down_proj_EFD, self.quantized_dtype, [0, 1], {},
                    "absmax")
            kernel_gating_EDF = kernel_gating_EDF.array
            kernel_up_proj_EDF = kernel_up_proj_EDF.array
            kernel_down_proj_EFD = kernel_down_proj_EFD.array

        return mapped_moe_fwd(self, x_TD, router_weights_TX,
                              selected_experts_TX, kernel_gating_EDF,
                              kernel_up_proj_EDF, kernel_down_proj_EFD)



import re
from typing import List, Optional, Tuple

import jax
import jax.numpy as jnp
from flax import nnx
from flax.typing import PRNGKey
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P
from vllm.config import VllmConfig

from tpu_inference.layers.jax.attention.attention import AttentionMetadata
from tpu_inference.layers.jax.attention.llama4_attention import Llama4Attention
from tpu_inference.layers.jax.constants import KVCacheType
from tpu_inference.layers.jax.layers import DenseFFW, Embedder, LMhead, RMSNorm
from tpu_inference.layers.jax.misc import shard_put
from tpu_inference.layers.jax.moe.moe import MoE, Router
from tpu_inference.layers.jax.transformer_block import \
    SharedExpertsTransformerBlock
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.utils.weight_utils import (
    convert_torch_to_jax_with_view, get_param, model_weights_generator,
    print_param_info, reshape_params, transpose_params)

logger = init_logger(__name__)


class Llama4ForCausalLM(nnx.Module):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: PRNGKey,
                 mesh: Mesh,
                 force_random_weights: bool = False):
        assert mesh is not None

        self.vllm_config = vllm_config
        model_config = vllm_config.model_config
        text_config = model_config.hf_config.text_config

        self.rng = nnx.Rngs(rng)
        self.mesh = mesh
        self.is_verbose = getattr(self.vllm_config.additional_config,
                                  "is_verbose", False)

        # Currently the runner will always set a mesh, so the custom default sharding (when
        #  no sharding is set in vllm config) doesn't take effect.
        # TODO(fhzhang): figure out whether we need to actually enable this.
        #    strategy_dict = {"tensor_parallelism": 4, "expert_parallelism": 2}

        self.vocab_size = model_config.get_vocab_size()
        self.hidden_size = model_config.get_hidden_size()

        dtype: jnp.dtype = jnp.bfloat16

        self.num_layers: int = getattr(text_config, "num_hidden_layers", 48)

        self.intermediate_size_moe: int = getattr(text_config,
                                                  "intermediate_size", 8192)
        self.intermediate_size_mlp = getattr(text_config,
                                             "intermediate_size_mlp", 16384)

        # num_local_experts: uses 16 experts for Llama-4-Scout-17B-16E-Instruct and uses 128 experts Llama-4-Maverick-17B-128E-Instruct.
        # The default value is set to 16 for compatibility with Llama-4-Scout.
        self.num_local_experts: int = getattr(text_config, "num_local_experts",
                                              16)
        self.hidden_act: str = getattr(text_config, "hidden_act", "silu")
        self.no_rope_layer_interval = 4

        # interleave_moe_layer_step has a layer step of 2 to interleave MoE and dense layers for Llama-4-Maverick-17B-128E-Instruct.
        # The default value is set to 1 for compatibility with Llama-4-Scout.
        self.interleave_moe_layer_step = getattr(text_config,
                                                 "interleave_moe_layer_step",
                                                 1)

        self.num_attention_heads = getattr(text_config, "num_attention_heads",
                                           40)
        self.num_key_value_heads = getattr(text_config, "num_key_value_heads",
                                           8)
        self.head_dim = getattr(text_config, "head_dim", 128)

        self.num_shared_experts = getattr(text_config, "num_experts_per_tok",
                                          1)
        self.rms_norm_eps = getattr(text_config, "rms_norm_eps", 1e-5)

        self.rope_scaling = getattr(text_config, "rope_scaling", None)
        if self.rope_scaling:
            self.rope_scaling["scale_factor"] = self.rope_scaling.pop("factor")

        self.use_qk_norm = getattr(text_config, "use_qk_norm", True)

        self.embedder = Embedder(vocab_size=self.vocab_size,
                                 hidden_size=self.hidden_size,
                                 dtype=dtype,
                                 vd_sharding=(('data', 'expert', 'model'),
                                              None),
                                 rngs=self.rng,
                                 random_init=force_random_weights)

        self.layers = []

        for i in range(self.num_layers):
            # For Llama4-Scout, all layers are MoE layers.
            # This can be adjusted for other variants.
            is_moe_layer = (i + 1) % \
                            self.interleave_moe_layer_step == 0

            # Llama-4-Scout config: It has "no_rope_layers": []
            use_attention_rope = (i + 1) % self.no_rope_layer_interval != 0

            router = Router(dtype=dtype,
                            hidden_size=self.hidden_size,
                            num_experts=self.num_local_experts,
                            num_experts_per_tok=1,
                            router_act="sigmoid",
                            rngs=self.rng,
                            activation_ffw_td=('data', None),
                            ed_sharding=(None, None),
                            random_init=force_random_weights)

            moe_ffw = MoE(
                dtype=dtype,
                num_local_experts=self.num_local_experts,
                apply_expert_weight_before_computation=True,
                hidden_size=self.hidden_size,
                intermediate_size_moe=self.intermediate_size_moe,
                hidden_act=self.hidden_act,
                router=router,
                rngs=self.rng,
                activation_ffw_td=('data', None),
                activation_ffw_ted=('data', 'expert', None),
                edf_sharding=('model', None, None),
                efd_sharding=('model', None, None),
                random_init=force_random_weights) if is_moe_layer else None

            dense_ffw = DenseFFW(
                dtype=dtype,
                hidden_act=self.hidden_act,
                hidden_size=self.hidden_size,
                intermediate_size=self.intermediate_size_mlp,
                random_init=force_random_weights,
                rngs=self.rng,
                df_sharding=(None, 'model'),
                fd_sharding=('model', None),
                activation_ffw_td=('data', None)) if not is_moe_layer else None

            attn = Llama4Attention(
                hidden_size=self.hidden_size,
                dtype=dtype,
                kv_cache_dtype=vllm_config.cache_config.cache_dtype,
                num_attention_heads=self.num_attention_heads,
                num_key_value_heads=self.num_key_value_heads,
                head_dim=self.head_dim,
                rope_theta=500000.0,
                # https://huggingface.co/meta-llama/Llama-4-Scout-17B-16E-Instruct/blob/main/config.json
                rope_scaling=self.rope_scaling,
                rngs=self.rng,
                rope_input_ordering="interleaved",
                temperature_tuning=True,
                temperature_tuning_scale=0.1,
                temperature_tuning_floor_scale=8192,
                use_qk_norm=self.use_qk_norm,
                attention_chunk_size=None if use_attention_rope else 8192,
                mesh=self.mesh,
                random_init=force_random_weights,
                activation_attention_td=('data', 'model'),
                activation_q_td=('data', 'model'),
                query_tnh=P('data', 'model', None),
                keyvalue_skh=P('data', 'model', None),
                activation_attention_out_td=('data', 'model'),
                attn_o_tnh=P('data', 'model', None),
                dnh_sharding=(None, 'model', None),
                dkh_sharding=(None, 'model', None),
                nhd_sharding=('model', None, None),
            )

            shared_experts = DenseFFW(
                dtype=dtype,
                hidden_act=self.hidden_act,
                hidden_size=self.hidden_size,
                intermediate_size=self.num_shared_experts *
                self.intermediate_size_moe,
                rngs=self.rng,
                random_init=force_random_weights,
                df_sharding=(None, 'model'),
                fd_sharding=('model', None),
                activation_ffw_td=('data', None)) if is_moe_layer else None

            pre_attention_norm = RMSNorm(
                dims=self.hidden_size,
                random_init=force_random_weights,
                epsilon=self.rms_norm_eps,
                rngs=self.rng,
                with_scale=True,
                dtype=dtype,
                activation_ffw_td=('data', None),
            )

            pre_mlp_norm = RMSNorm(
                dims=self.hidden_size,
                epsilon=self.rms_norm_eps,
                rngs=self.rng,
                with_scale=True,
                dtype=dtype,
                random_init=force_random_weights,
                activation_ffw_td=('data', None),
            )

            block = SharedExpertsTransformerBlock(
                moe_ffw=moe_ffw if is_moe_layer else None,
                dense_ffw=dense_ffw if not is_moe_layer else None,
                shared_experts=shared_experts if is_moe_layer else None,
                attn=attn,
                pre_attention_norm=pre_attention_norm,
                pre_mlp_norm=pre_mlp_norm,
                use_attention_rope=use_attention_rope)
            self.layers.append(block)

        self.final_norm = RMSNorm(
            dims=self.hidden_size,
            epsilon=self.rms_norm_eps,
            rngs=self.rng,
            with_scale=True,
            dtype=dtype,
            random_init=force_random_weights,
        )

        self.lm_head = LMhead(vocab_size=self.vocab_size,
                              hidden_size=self.hidden_size,
                              dtype=dtype,
                              rngs=self.rng,
                              vd_sharding=(('data', 'expert', 'model'), None),
                              dv_sharding=(None, ('data', 'expert', 'model')),
                              random_init=force_random_weights)
        if self.is_verbose:
            self._print_model_architecture()

    def _print_model_architecture(self):
        num_display_layers = max(self.interleave_moe_layer_step,
                                 self.no_rope_layer_interval)

        logger.info("### Embedding ###")
        nnx.display(self.embedder)

        logger.info(f"\n### First {num_display_layers} Layers ###")
        # Loop through the slice and display each layer
        for i, layer in enumerate(self.layers[:num_display_layers]):
            logger.info(f"\n--- Layer {i} ---")
            nnx.display(layer)

        logger.info("\n### LM Head ###")
        nnx.display(self.lm_head)

    def load_weights(self, rng: jax.Array, cache_dir: Optional[str] = None):
        # NOTE: Since we are using nnx.eval_shape to init the model,
        # we have to pass dynamic arrays here for __call__'s usage.
        self.rng = nnx.Rngs(rng)

        weight_loader = Llama4WeightLoader(
            vllm_config=self.vllm_config,
            hidden_size=self.hidden_size,
            attn_heads=self.num_attention_heads,
            num_key_value_heads=self.num_key_value_heads,
            attn_head_dim=self.head_dim)
        weight_loader.load_weights(self)

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        *args,
    ) -> Tuple[List[KVCacheType], jax.Array, List[jax.Array]]:
        is_prefill = False
        x_TD = self.embedder.encode(input_ids)

        for (i, block) in enumerate(self.layers):
            kv_cache = kv_caches[i]
            new_kv_cache, x_TD = block(x_TD, is_prefill, kv_cache,
                                       attention_metadata)
            jax.block_until_ready(x_TD)
            kv_caches[i] = new_kv_cache

        final_activation_TD = self.final_norm(x_TD)

        return kv_caches, final_activation_TD, []

    def compute_logits(self, hidden_states: jax.Array) -> jax.Array:
        logits_TV = jnp.dot(hidden_states,
                            self.lm_head.input_embedding_table_DV.value)
        return logits_TV


class Llama4WeightLoader:

    def __init__(self, vllm_config: VllmConfig, hidden_size, attn_heads,
                 num_key_value_heads, attn_head_dim):
        self.names_and_weights_generator = model_weights_generator(
            model_name_or_path=vllm_config.model_config.model,
            framework="pt",
            filter_regex="language_model",
            download_dir=vllm_config.load_config.download_dir)
        self.is_verbose = getattr(vllm_config.additional_config, "is_verbose",
                                  False)
        self.interleave_moe_layer_step = getattr(
            vllm_config.model_config.hf_config.text_config,
            "interleave_moe_layer_step", 1)

        self.quantization_config = getattr(vllm_config.model_config.hf_config,
                                           "quantization_config", None)
        self.expert_weights_buffer = {}
        self.expert_prefix = "shared_expert."

        transpose_mappings_to_quantization = {
            "down_proj": (1, 0),
            "gate_proj": (1, 0),
            "up_proj": (1, 0),
        }

        self._transpose_map = {
            "q_proj": (2, 0, 1),
            "k_proj": (2, 0, 1),
            "v_proj": (2, 0, 1),
            "router": (1, 0),
            f"{self.expert_prefix}down_proj": (1, 0),
            f"{self.expert_prefix}gate_proj": (1, 0),
            f"{self.expert_prefix}up_proj": (1, 0),
            "feed_forward.down_proj": (1, 0),
            "feed_forward.gate_proj": (1, 0),
            "feed_forward.up_proj": (1, 0),
            "o_proj": (1, 2, 0),
            "lm_head": (1, 0),
        }

        if self.quantization_config and self.expert_prefix:
            self._transpose_map.update(transpose_mappings_to_quantization)

        self._weight_shape_map = {
            "q_proj": (attn_heads, attn_head_dim, hidden_size),
            "k_proj": (num_key_value_heads, attn_head_dim, hidden_size),
            "v_proj": (num_key_value_heads, attn_head_dim, hidden_size),
            # o_proj is inverted: https://github.com/huggingface/transformers/blob/v4.53.2/src/transformers/models/llama4/modeling_llama4.py#L298
            "o_proj": (hidden_size, attn_heads, attn_head_dim),
        }

        # Set the mappings from loaded parameter keys to standardized names.\
        # 1. EXPERT_MAPPINGS_FUSED: Used for non-quantized (e.g., BF16) checkpoints.
        #    - This format typically comes from standard checkpoints where 'gate' and 'up' projection weights might be combined (FUSED) into a single tensor.
        #    - Expert weights are usually stacked, with the expert dimension (E) being the first dimension.
        EXPERT_MAPPINGS_FUSED = {
            "language_model.model.layers.*.feed_forward.experts.down_proj":
            "layers.*.moe_ffw.kernel_down_proj_EFD",
            "language_model.model.layers.*.feed_forward.experts.gate_up_proj":
            "layers.*.moe_ffw.kernel_up_proj_EDF",
        }

        # 2. EXPERT_MAPPINGS_UNFUSED: Specifically designed for quantized checkpoints (e.g., FP8).
        #    - Quantized checkpoints store each expert's weights separately and explicitly separate the 'weight' (quantized value) from the 'weight_scale' (quantization scale).
        #    - The mapping captures both the `.weight` and `.weight_scale` components. This allows the loader to aggregate (stack) the individual expert weights and scales.
        EXPERT_MAPPINGS_UNFUSED = {
            "language_model.model.layers.*.feed_forward.experts.*.down_proj.weight":
            "layers.*.moe_ffw.kernel_down_proj_EFD",
            "language_model.model.layers.*.feed_forward.experts.*.down_proj.weight_scale":
            "layers.*.moe_ffw.kernel_down_proj_EFD",
            "language_model.model.layers.*.feed_forward.experts.*.gate_proj.weight":
            "layers.*.moe_ffw.kernel_gating_EDF",
            "language_model.model.layers.*.feed_forward.experts.*.gate_proj.weight_scale":
            "layers.*.moe_ffw.kernel_gating_EDF",
            "language_model.model.layers.*.feed_forward.experts.*.up_proj.weight":
            "layers.*.moe_ffw.kernel_up_proj_EDF",
            "language_model.model.layers.*.feed_forward.experts.*.up_proj.weight_scale":
            "layers.*.moe_ffw.kernel_up_proj_EDF",
        }

        self._loaded_to_standardized_keys = {
            "language_model.model.embed_tokens.weight":
            "embedder.input_embedding_table_VD",
            "language_model.lm_head.weight":
            "lm_head.input_embedding_table_DV",
            "language_model.model.norm.weight":
            "final_norm.scale",
            "language_model.model.layers.*.input_layernorm.weight":
            "layers.*.pre_attention_norm.scale",
            "language_model.model.layers.*.post_attention_layernorm.weight":
            "layers.*.pre_mlp_norm.scale",
            "language_model.model.layers.*.self_attn.q_proj.weight":
            "layers.*.attn.kernel_q_proj_DNH",
            "language_model.model.layers.*.self_attn.k_proj.weight":
            "layers.*.attn.kernel_k_proj_DKH",
            "language_model.model.layers.*.self_attn.v_proj.weight":
            "layers.*.attn.kernel_v_proj_DKH",
            "language_model.model.layers.*.self_attn.o_proj.weight":
            "layers.*.attn.kernel_o_proj_NHD",
            "language_model.model.layers.*.feed_forward.router.weight":
            "layers.*.moe_ffw.router.kernel_DE",
            # shared experts
            "language_model.model.layers.*.feed_forward.shared_expert.down_proj.weight":
            "layers.*.shared_experts.kernel_down_proj_FD",
            "language_model.model.layers.*.feed_forward.shared_expert.gate_proj.weight":
            "layers.*.shared_experts.kernel_gating_DF",
            "language_model.model.layers.*.feed_forward.shared_expert.up_proj.weight":
            "layers.*.shared_experts.kernel_up_proj_DF",
            # dense layers
            "language_model.model.layers.*.feed_forward.down_proj.weight":
            "layers.*.dense_ffw.kernel_down_proj_FD",
            "language_model.model.layers.*.feed_forward.up_proj.weight":
            "layers.*.dense_ffw.kernel_up_proj_DF",
            "language_model.model.layers.*.feed_forward.gate_proj.weight":
            "layers.*.dense_ffw.kernel_gating_DF",
        }

        if self.quantization_config is None:
            self._loaded_to_standardized_keys.update(EXPERT_MAPPINGS_FUSED)
        else:
            self._loaded_to_standardized_keys.update(EXPERT_MAPPINGS_UNFUSED)

    def map_loaded_to_standardized_name(self, loaded_key: str) -> str:
        # Find the corresponding model key using the HF key
        if "layer" in loaded_key:
            layer_num = self._get_layer_num(loaded_key)
            layer_key = re.sub(r"layers\.\d+", "layers.*", loaded_key)

            expert_match = re.search(r"experts\.(\d+)", layer_key)
            if expert_match:
                # Key for lookup eg: layers.*.feed_forward.experts.*.down_proj.weight
                layer_key = re.sub(r"experts\.\d+", "experts.*", layer_key)

            mapped_key = self._loaded_to_standardized_keys.get(
                layer_key, loaded_key)
            mapped_key = re.sub(r"layers\.\*", f"layers.{layer_num}",
                                mapped_key)
        else:
            mapped_key = self._loaded_to_standardized_keys.get(
                loaded_key, loaded_key)
        return mapped_key

    def _map_llama4_gate_up_proj(self, model_for_loading: nnx.Module,
                                 model_params: nnx.State, loaded_name: str,
                                 loaded_weight: jax.Array):
        """HF's gate_up_proj is a fused tensor of gate and up projections. It needs to be split."""

        cast_type = jnp.dtype(jnp.bfloat16)
        # loaded_weight is a jax.Array when framework="flax", otherwise it's bfloat16
        if not isinstance(loaded_weight, jax.Array):
            loaded_weight = convert_torch_to_jax_with_view(
                loaded_weight, cast_type)

        split_weights = jnp.split(loaded_weight, 2, axis=-1)
        layer_num = self._get_layer_num(loaded_name)

        for split_type in ["gate", "up"]:
            split_loaded_name = loaded_name.replace("gate_up_proj",
                                                    f"{split_type}_proj")
            if split_type == "gate":
                mapped_name = "layers.*.moe_ffw.kernel_gating_EDF"
                loaded_weight = split_weights[0]
            else:
                mapped_name = "layers.*.moe_ffw.kernel_up_proj_EDF"
                loaded_weight = split_weights[1]

            mapped_name = re.sub(r"layers\.\*", f"layers.{layer_num}",
                                 mapped_name)

            mapped_model_weight = get_param(model_params, mapped_name)

            if mapped_model_weight.value.shape != loaded_weight.shape:
                raise ValueError(
                    f"Loaded shape for {split_loaded_name}: {loaded_weight.shape} "
                    f"does not match model shape for {mapped_name}: {mapped_model_weight.value.shape}!"
                )

            mapped_model_weight.value = shard_put(loaded_weight,
                                                  mapped_model_weight.sharding,
                                                  mesh=model_for_loading.mesh)
            logger.debug(
                f"{split_loaded_name}: {loaded_weight.shape}  -->  {mapped_name}: {mapped_model_weight.value.shape}"
            )
            if self.is_verbose:
                print_param_info(mapped_model_weight, mapped_name)

    def _get_layer_num(self, loaded_key: str) -> Optional[int]:
        """
        Extracts the layer number from a HuggingFace weight key string.
        Returns the layer number (int) or None if no layer number is found.
        """
        match = re.search(r"layers\.(\d+)", loaded_key)
        if match:
            return int(match.group(1))
        return None

    def _get_expert_num(self, loaded_key: str) -> Optional[int]:
        """
        Extracts the expect number from a HuggingFace weight key string.
        Returns the expect number (int) or None if no expect number is found.
        """
        match = re.search(r"experts\.(\d+)\.", loaded_key)
        if match:
            return int(match.group(1))
        return None

    def load_weights(self, model_for_loading: nnx.Module):
        model_params = nnx.state(model_for_loading)

        with jax.default_device(jax.devices("cpu")[0]):
            for loaded_name, loaded_weight in self.names_and_weights_generator:
                is_moe_layer = False
                layer_num = self._get_layer_num(loaded_name)
                expert_num = self._get_expert_num(loaded_name)
                # Quantized (FP8) checkpoints unstack the expert weights, while unquantized (BF16) checkpoints keep them stacked.
                is_unfused_expert = self.quantization_config is not None and expert_num is not None
                is_scale = loaded_name.endswith(".weight_scale")

                if is_unfused_expert:
                    mapped_name = self.map_loaded_to_standardized_name(
                        loaded_name)
                    model_weight = get_param(model_params, mapped_name)

                    if is_scale:
                        cast_type = model_weight.array.scale.value.dtype
                    else:
                        cast_type = model_weight.array.qvalue.value.dtype

                    loaded_weight = convert_torch_to_jax_with_view(
                        loaded_weight, cast_type)
                    loaded_weight = transpose_params(loaded_name,
                                                     loaded_weight,
                                                     self._transpose_map)

                    buffer_key = f"{mapped_name}_{'scale' if is_scale else 'qvalue'}"
                    if buffer_key not in self.expert_weights_buffer:
                        self.expert_weights_buffer[buffer_key] = {}
                    self.expert_weights_buffer[buffer_key][
                        expert_num] = loaded_weight
                    continue

                if layer_num is not None:
                    is_moe_layer = (layer_num + 1) % \
                            self.interleave_moe_layer_step == 0
                    self.expert_prefix = "shared_expert." if is_moe_layer else ""

                if "gate_up_proj" in loaded_name:
                    self._map_llama4_gate_up_proj(model_for_loading,
                                                  model_params, loaded_name,
                                                  loaded_weight)
                    continue

                mapped_name = self.map_loaded_to_standardized_name(loaded_name)
                model_weight = get_param(model_params, mapped_name)

                cast_type = model_weight.value.dtype
                if not isinstance(loaded_weight, jax.Array):
                    logger.debug(
                        f"Converting PyTorch tensor {loaded_name} to JAX {cast_type}"
                    )
                    loaded_weight = convert_torch_to_jax_with_view(
                        loaded_weight, cast_type)

                if not loaded_name.endswith(".bias"):
                    loaded_weight = reshape_params(loaded_name, loaded_weight,
                                                   self._weight_shape_map)
                    loaded_weight = transpose_params(loaded_name,
                                                     loaded_weight,
                                                     self._transpose_map)
                if model_weight.value.shape != loaded_weight.shape:
                    raise ValueError(
                        f"Loaded shape for {loaded_name}: {loaded_weight.shape} "
                        f"does not match model shape for {mapped_name}: {model_weight.value.shape}!"
                    )
                logger.debug(
                    f"Transformed parameter {loaded_name} to {mapped_name}: {loaded_weight.shape} --> {model_weight.value.shape}"
                )

                model_weight.value = shard_put(loaded_weight,
                                               model_weight.sharding,
                                               mesh=model_for_loading.mesh)
                if self.is_verbose:
                    print_param_info(model_weight, loaded_name)

            with jax.default_device(jax.devices("cpu")[0]):
                for buffer_key, expert_map in self.expert_weights_buffer.items(
                ):
                    sorted_exp_nums = sorted(expert_map.keys())
                    aggregated_weight = jnp.stack(
                        [expert_map[k] for k in sorted_exp_nums], axis=0)
                    is_scale = buffer_key.endswith("_scale")
                    base_mapped_name = buffer_key.replace("_scale",
                                                          "").replace(
                                                              "_qvalue", "")

                    model_weight = get_param(model_params, base_mapped_name)

                    assert hasattr(
                        model_weight, 'array'
                    ), f"Expected MoE weight '{base_mapped_name}' to be a quantized array (qarray)"

                    if is_scale:
                        loaded_name = f"{base_mapped_name}.array.scale.value"
                        if model_weight.array.scale.value.shape != aggregated_weight.shape:
                            raise ValueError(
                                f"[AGGREGATED] Loaded shape for {buffer_key}: {aggregated_weight.shape}"
                                f"does not match model shape for {loaded_name}: {model_weight.array.scale.value.shape}!"
                            )

                        model_weight.array.scale.value = shard_put(
                            aggregated_weight,
                            model_weight.array.scale.sharding,
                            mesh=model_for_loading.mesh)

                    elif aggregated_weight.itemsize < 2:  # check model weight elem nbits < 16
                        loaded_name = f"{base_mapped_name}.array.qvalue.value"
                        if model_weight.array.qvalue.value.shape != aggregated_weight.shape:
                            raise ValueError(
                                f"[AGGREGATED] Loaded shape for {buffer_key}: {aggregated_weight.shape}"
                                f"does not match model shape for {loaded_name}: {model_weight.array.qvalue.value.shape}!"
                            )

                        model_weight.array.qvalue.value = shard_put(
                            aggregated_weight,
                            model_weight.array.qvalue.sharding,
                            mesh=model_for_loading.mesh)

                    logger.debug(
                        f"Aggregated and loaded {loaded_name}: {aggregated_weight.shape}"
                    )

                    if self.is_verbose:
                        print_param_info(model_weight, loaded_name)

        nnx.update(model_for_loading, model_params)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from torchax.interop import jax_view
from vllm.lora.layers.base_linear import BaseLinearLayerWithLoRA
from vllm.lora.request import LoRARequest

from tpu_inference.layers.vllm.sharding import update_lora

if TYPE_CHECKING:
    from tpu_inference.runner.tpu_runner import TPUModelRunner


class LoraUtils:

    def __init__(self, runner: "TPUModelRunner"):
        self.runner = runner

    def set_active_loras(self, num_scheduled_tokens_per_req,
                         total_num_scheduled_tokens,
                         padded_total_num_scheduled_tokens):
        # We need to respect padding when activating LoRA adapters
        padded_num_scheduled_tokens_per_req = np.copy(
            num_scheduled_tokens_per_req
        )  # Copying to avoid accidental state corruption bugs
        padded_num_scheduled_tokens_per_req[-1] += \
            padded_total_num_scheduled_tokens - total_num_scheduled_tokens

        prompt_lora_mapping: tuple[int, ...]  # of size input_batch.num_reqs
        token_lora_mapping: tuple[int,
                                  ...]  # of size np.sum(num_scheduled_tokens)
        lora_requests: set[LoRARequest]
        prompt_lora_mapping, token_lora_mapping, lora_requests = \
                            self.runner.input_batch.make_lora_inputs(padded_num_scheduled_tokens_per_req)
        # One should not put lora_manager.set_active_loras under
        # torchax.default_env() because set_active_loras also load lora from
        # disk and torchax currently does not support that.  Here we load the
        # lora and set the lora weight to the linear layers.
        self.runner._set_active_loras(prompt_lora_mapping, token_lora_mapping,
                                      lora_requests)

        params_and_buffers = update_lora(
            self.runner.model.model, initial_params_buffers=self.runner.state)
        self.runner.state = jax_view(params_and_buffers)

    def extract_lora_metadata(self):
        if self.runner.lora_config is None:
            return None

        metadata = {}
        punica_wrapper = None
        for _, m in self.runner.model.model.named_modules():
            if isinstance(m, BaseLinearLayerWithLoRA):
                assert getattr(
                    m, 'punica_wrapper', None
                ) is not None, 'A lora wrapper should have contained a punica_wrapper'
                punica_wrapper = m.punica_wrapper
                break
        assert punica_wrapper is not None, 'Should have been able to find a punica wrapper from the Lora wrapper.'

        # vars does not show inherited methods or class attributes but this is
        # fine because we only care about instance attributes.
        for k in vars(punica_wrapper):
            v = getattr(punica_wrapper, k, None)
            if k == 'device':  # Exclude string as it can't be traced by jax.jit
                continue
            metadata[k] = v
        return jax_view(metadata)


def replace_lora_metadata(model, metadata: dict, lora_config) -> dict:
    if lora_config is None or not metadata:
        return {}

    original_metadata = {}
    punica_wrapper = None
    for _, m in model.named_modules():
        if isinstance(m, BaseLinearLayerWithLoRA):
            assert getattr(
                m, 'punica_wrapper', None
            ) is not None, 'A lora wrapper should have contained a punica_wrapper'
            punica_wrapper = m.punica_wrapper
            break
    assert punica_wrapper is not None, 'Should have been able to find a punica wrapper from the Lora wrapper.'

    for k in vars(punica_wrapper):
        if k in metadata:
            original_metadata[k] = getattr(punica_wrapper, k)
            setattr(punica_wrapper, k, metadata[k])
    return original_metadata



# TODO: Update documentation

from typing import List, Optional, Tuple

import jax
import jax.numpy as jnp
from flax import nnx
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P
from vllm.config import VllmConfig

from tpu_inference.layers.jax.attention.attention import (Attention,
                                                          AttentionMetadata)
from tpu_inference.layers.jax.constants import KVCacheType
from tpu_inference.layers.jax.layers import DenseFFW, Embedder, LMhead, RMSNorm
from tpu_inference.layers.jax.transformer_block import TransformerBlock
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.utils.weight_utils import (MetadataMap,
                                                         load_hf_weights)

logger = init_logger(__name__)


class LlamaForCausalLM(nnx.Module):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: jax.Array,
                 mesh: Mesh,
                 force_random_weights: bool = False):
        assert mesh is not None

        self.vllm_config = vllm_config
        self.rng = nnx.Rngs(rng)
        self.mesh = mesh

        model_name = self.vllm_config.model_config.model.lower()
        if "70b" in model_name:
            logger.info("Initializing Llama3 70B model variant.")
            self.hidden_size = 8192
            num_layers = 80
            self.num_attention_heads = 64
            self.num_key_value_heads = 8
            intermediate_size = 28672
        elif "8b" in model_name:
            logger.info("Initializing Llama3 8B model variant.")
            self.hidden_size = 4096
            num_layers = 32
            self.num_attention_heads = 32
            self.num_key_value_heads = 8
            intermediate_size = 14336
        else:
            raise ValueError(
                f"Could not determine Llama3 variant (8B or 70B) from model name: '{model_name}'. "
                "Please ensure '8b' or '70b' is in the model path.")

        dtype = jnp.bfloat16
        self.head_dim = 128
        rope_theta = 500000.0
        vocab_size = 128256
        rms_norm_eps = 1e-5

        self.embedder = Embedder(vocab_size=vocab_size,
                                 hidden_size=self.hidden_size,
                                 dtype=dtype,
                                 rngs=self.rng,
                                 random_init=force_random_weights,
                                 vd_sharding=("model", None))

        self.layers = []
        kv_cache_dtype = self.vllm_config.cache_config.cache_dtype
        for _ in range(num_layers):
            self.layers.append(
                TransformerBlock(
                    pre_attention_norm=RMSNorm(
                        dims=self.hidden_size,
                        random_init=force_random_weights,
                        epsilon=rms_norm_eps,
                        rngs=self.rng,
                        with_scale=True,
                        dtype=dtype,
                    ),
                    pre_mlp_norm=RMSNorm(
                        dims=self.hidden_size,
                        rngs=self.rng,
                        random_init=force_random_weights,
                        epsilon=rms_norm_eps,
                        with_scale=True,
                        dtype=dtype,
                    ),
                    attn=Attention(
                        hidden_size=self.hidden_size,
                        num_attention_heads=self.num_attention_heads,
                        num_key_value_heads=self.num_key_value_heads,
                        head_dim=self.head_dim,
                        rope_theta=rope_theta,
                        rope_scaling={},
                        rngs=self.rng,
                        dtype=dtype,
                        # TODO (jacobplatin): we should refactor this to pass a dtype (or config) directly
                        kv_cache_dtype=kv_cache_dtype,
                        mesh=self.mesh,
                        random_init=force_random_weights,
                        dnh_sharding=(None, "model", None),
                        dkh_sharding=(None, "model", None),
                        nhd_sharding=("model", None, None),
                        query_tnh=P(None, "model", None),
                        keyvalue_skh=P(None, "model", None),
                        attn_o_tnh=P(None, "model", None),
                    ),
                    custom_module=DenseFFW(dtype=dtype,
                                           hidden_act="silu",
                                           hidden_size=self.hidden_size,
                                           intermediate_size=intermediate_size,
                                           rngs=self.rng,
                                           df_sharding=(None, "model"),
                                           fd_sharding=("model", None),
                                           random_init=force_random_weights),
                ))

        self.final_norm = RMSNorm(
            dims=self.hidden_size,
            rngs=self.rng,
            random_init=force_random_weights,
            epsilon=rms_norm_eps,
            with_scale=True,
            dtype=dtype,
        )

        self.lm_head = LMhead(vocab_size=vocab_size,
                              hidden_size=self.hidden_size,
                              dtype=dtype,
                              rngs=self.rng,
                              dv_sharding=(None, 'model'),
                              random_init=force_random_weights)

    def load_weights(self, rng: jax.Array, cache_dir: Optional[str] = None):
        # NOTE: Since we are using nnx.eval_shape to init the model,
        # we have to pass dynamic arrays here for __call__'s usage.
        self.rng = nnx.Rngs(rng)
        weight_loader = Llama3WeightLoader(
            vllm_config=self.vllm_config,
            hidden_size=self.hidden_size,
            attn_heads=self.num_attention_heads,
            num_key_value_heads=self.num_key_value_heads,
            attn_head_dim=self.head_dim)

        weight_loader.load_weights(self)

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        *args,
    ) -> Tuple[List[KVCacheType], jax.Array]:
        is_prefill = False
        with jax.named_scope("llama_embed_input"):  #Embedding
            x_TD = self.embedder.encode(input_ids)

        with jax.named_scope("llama_model_transformer_blocks"):
            for (i, layer) in enumerate(self.layers):
                kv_cache = kv_caches[i]

                # The first layer is unscoped to avoid JAX tracing issues.
                # JAX's profiler may incorrectly apply the scope name from the first
                # layer's kernel compilation to all subsequent layers. Skipping the
                # first layer ensures distinct scope names for the remaining layers.
                if i == 0:
                    new_kv_cache, x_TD = layer(x_TD, is_prefill, kv_cache,
                                               attention_metadata)
                else:
                    with jax.named_scope(f'layer_{i}'):
                        new_kv_cache, x_TD = layer(x_TD, is_prefill, kv_cache,
                                                   attention_metadata)

                kv_caches[i] = new_kv_cache

        with jax.named_scope(
                "llama_final_norm"):  #Norm after last transformer block
            final_activation_TD = self.final_norm(x_TD)

        return kv_caches, final_activation_TD, []

    def compute_logits(self, hidden_states: jax.Array) -> jax.Array:
        with jax.named_scope("llama_lm_head_projection"
                             ):  #LM head projection to produce logits
            logits_TV = jnp.dot(hidden_states,
                                self.lm_head.input_embedding_table_DV.value)

        return logits_TV


class Llama3WeightLoader:

    def __init__(self, vllm_config: VllmConfig, hidden_size, attn_heads,
                 num_key_value_heads, attn_head_dim):
        self._transpose_map = {
            "lm_head": (1, 0),
            "gate_proj": (1, 0),
            "up_proj": (1, 0),
            "down_proj": (1, 0),
            "q_proj": (2, 0, 1),
            "k_proj": (2, 0, 1),
            "v_proj": (2, 0, 1),
            "o_proj": (1, 2, 0),
        }
        self._weight_shape_map = {
            "q_proj": (attn_heads, -1, hidden_size),
            "k_proj": (num_key_value_heads, -1, hidden_size),
            "v_proj": (num_key_value_heads, -1, hidden_size),
            "o_proj": (hidden_size, attn_heads, -1),
        }
        self._bias_shape_map = {
            "q_proj.bias": (attn_heads, attn_head_dim),
            "k_proj.bias": (num_key_value_heads, attn_head_dim),
            "v_proj.bias": (num_key_value_heads, attn_head_dim)
        }

        # Set the mappings from loaded parameter keys to standardized names.
        self._loaded_to_standardized_keys = {
            "model.embed_tokens": "embedder.input_embedding_table_VD",
            "model.layers.*.input_layernorm":
            "layers.*.pre_attention_norm.scale",
            "model.layers.*.mlp.down_proj":
            "layers.*.custom_module.kernel_down_proj_FD",
            "model.layers.*.mlp.gate_proj":
            "layers.*.custom_module.kernel_gating_DF",
            "model.layers.*.mlp.up_proj":
            "layers.*.custom_module.kernel_up_proj_DF",
            "model.layers.*.post_attention_layernorm":
            "layers.*.pre_mlp_norm.scale",
            "model.layers.*.self_attn.k_proj":
            "layers.*.attn.kernel_k_proj_DKH",
            "model.layers.*.self_attn.o_proj":
            "layers.*.attn.kernel_o_proj_NHD",
            "model.layers.*.self_attn.q_proj":
            "layers.*.attn.kernel_q_proj_DNH",
            "model.layers.*.self_attn.v_proj":
            "layers.*.attn.kernel_v_proj_DKH",
            "model.norm": "final_norm.scale",
            "lm_head": "lm_head.input_embedding_table_DV"
        }
        self.vllm_config = vllm_config

    def load_weights(self, model_for_loading: nnx.Module):
        model_params = nnx.state(model_for_loading)
        metadata_map = MetadataMap(name_map=self._loaded_to_standardized_keys,
                                   reshape_map=self._weight_shape_map,
                                   bias_reshape_map=self._bias_shape_map,
                                   transpose_map=self._transpose_map)
        load_hf_weights(vllm_config=self.vllm_config,
                        model=model_for_loading,
                        metadata_map=metadata_map,
                        mesh=model_for_loading.mesh)

        # TODO: validate that all of the model_params were accounted for as well.
        nnx.update(model_for_loading, model_params)
